#
# IBM Confidential
# OCO Source Materials
# Tivoli Access Manager for Operating Systems - Agent
# TIVO-TAMOSA
#
# (C) Copyright IBM Corp. 2008
# The source code for this program is not published or otherwise
# divested of its trade secrets, irrespective of what has been
# deposited with the U.S. Copyright Office.
#

#
# aix file audit implementation
#

from com.ibm.amos.agent.db import PolicyDelta
from com.ibm.amos.agent.db import PolicyConstants
from com.ibm.amos.agent.db import ResourceQuery
from com.ibm.amos.agent.db.resources import File
from com.ibm.amos.agent.db.rules import Audit

from ossutil.aixaudit import *
from ossutil import wrapPyException
from com.ibm.amos.nls.pdopmidx import *

import java.lang.StringBuffer
import java.lang.Exception
import java.lang.Integer
import re


class AIXFileAuditPlugin:
    """
    Handles AIX file audit updates.
    """

    def __init__(self, realplugin, target):
        self.realplugin = realplugin
        self.target = target

    def shutdown(self):
        pass


    # Get the list of entries (filenames) to remove from the audit
    # configuration, based on a policy delta.
    def getEntriesToRemove(self, delta):
       removeList = []

       # Remove the audit entry if a non-valid audit setting has been applied,
       # the file has been deleted from the objectspace, or the attached audit
       # has been changed to be invalid.
       rq = ResourceQuery(PolicyConstants.FILE)
       rq.resourceQuery(PolicyDelta.CHANGED)
       rq.resourceQuery(PolicyDelta.DELETED)
       rq.resourceRuleQuery(PolicyDelta.CHANGED, PolicyConstants.AUDIT)

       changedFiles = rq.evaluate(delta)
       if changedFiles is not None:
          for fileObj in changedFiles:
             file = fileObj.getKey()
             if isValidAuditOnFile(file, delta.oldPolicy) and \
                not isValidAuditOnFile(file, delta.newPolicy):

                removeList.append(file)

       return removeList
    
    def policyUpdate(self, result, delta, apply):

        try:
            auditObjFile = AIXAuditObjectsFile(result)
            outcome = npm_s_success

            xac = self.target.translatePath(AUDIT_OBJECTS)
            data = java.lang.String(self.target.getFile(xac))

            if not auditObjFile.parseFile(data):
                result.warning("Error parsing audit objects file");

            if delta.newPolicy != None:
                if not auditObjFile.buildEntries(delta.newPolicy):
                    outcome = npm_s_policy_not_compliant

            # Remove the entries if either an audit POP has been detached,
            # or the object has been deleted and it previously had an audit
            # POP attached.
            removeList = self.getEntriesToRemove(delta)
            if not auditObjFile.deleteEntries(removeList):
               outcome = npm_s_policy_not_compliant

            newData = auditObjFile.getFile()
            if apply and data.equals(newData) is False:
                result.startAction(File(AUDIT_OBJECTS, None, None))
                xac = self.target.translatePath(AUDIT_OBJECTS)
                self.target.putFile(xac, newData, True)
                restartAuditDaemon(self.target, result)
                result.finishAction(npm_s_success, [])

            result.finish(outcome, [])
        except java.lang.Exception, e:
            result.finish(e)
        except Exception, e:
            result.finish(wrapPyException(self.realplugin, e))

    def verifyPolicy(self, result, policy, reconcile):

        try:
            auditObjFile = AIXAuditObjectsFile(result)

            outcome = npm_s_success

            xac = self.target.translatePath(AUDIT_OBJECTS)
            data = java.lang.String(self.target.getFile(xac))

            rebuild = True
            if auditObjFile.parseFile(data):
                if auditObjFile.matchPolicy(policy, reconcile):
                    rebuild = False
            else:
                result.warning("Error parsing audit objects file")

            if rebuild:
                outcome = npm_s_policy_not_compliant
                if reconcile:
                    outcome = npm_s_policy_reconciled
                    newData = auditObjFile.getFile()
                    if newData != data:
                        result.startAction(File(AUDIT_OBJECTS, None, None))
                        self.target.putFile(xac, auditObjFile.getFile(), True)
                        restartAuditDaemon(self.target, result)
                        result.finishAction(npm_s_success, [])
            
            result.finish(outcome, [])
        except java.lang.Exception, e:
            result.finish(e)
        except Exception, e:
            result.finish(wrapPyException(self.realplugin, e))


class AIXAuditObjectsFile:
    """
    Represents the 'objects' file on AIX. Converts policy into entries
    for this file.
    """
    
    def __init__(self, result):
        self.result = result
        self.objects = {}
        self.objectsList = []

    def deleteEntry(self, item):
       if item in self.objects:
          for i in range(self.objects[item].arrPos + 1, len(self.objectsList)):
             self.objectsList[i].arrPos -= 1
          del self.objectsList[self.objects[item].arrPos]
          del self.objects[item]
       return True

    def deleteEntries(self, removeList):
       for item in removeList:
          self.deleteEntry(item)
       return True
       

    def buildEntries(self, policy):
        if policy.resourceMap is None or policy.ruleMap is None:
            # no valid audit policy can possibly exist
            return True

        files = policy.resourceMap.get(PolicyConstants.FILE)
        audits = policy.ruleMap.get(PolicyConstants.AUDIT)
        if files is None or audits is None:
            # no valid audit policy exists
            return True

        result = True
        for filename in files:
            fileObj = files.get(filename)
            if fileObj.getAuditID() is None:
                continue

            auditObj = audits.get(fileObj.getAuditID())

            self.result.startAction(fileObj)
            self.result.addRule(auditObj)
            self.addObject(fileObj, auditObj)
            self.result.finishAction(npm_s_success, [])

        return result


    def parseFile(self, data): 
        re_obj = re.compile("^\s*(/[/\w\.-]*):\s*([#\*].*)*$")              # Match object lines
        re_empty = re.compile("^\s*([rwx])\s*=\s*$")                        # Match against empty value, eg. r = 
        re_quote = re.compile("^\s*([rwx])\s*=\s*\"(\w*)\"\s*$")            # Match against quoted value eg. r = ""
        re_plain = re.compile("^\s*([rwx])\s*=\s*(\w+)\s*$")                # Match against plain value  eg. r = blah
        re_com = re.compile("^\s*[#\*].*$")                                 # Match comment lines
        lines = data.split("\n")
        currentObj = ""
        success = True
        for index, line in enumerate(lines):
            line.strip()

            mode = ""
            auditcls = ""
            addEvent = False
            
            if line == "" or re_com.match(line):
                continue
            elif re_obj.match(line):
                grp_obj = re_obj.match(line)
                currentObj = grp_obj.group(1)
                if currentObj in self.objects:
                    self.result.warning("Duplicate object '" + currentObj + "' encountered. Line: " + java.lang.Integer.toString(index+1))
                    success = False
                else:
                    self.objects[currentObj] = AIXAuditObject(currentObj, self.result)
                    self.objects[currentObj].arrPos = len(self.objectsList)
                    self.objectsList.append(self.objects[currentObj])

            elif re_quote.match(line):
               grp = re_quote.match(line)
               mode = grp.group(1)
               auditcls = grp.group(2)
               addEvent = True
            elif re_plain.match(line):
               grp = re_plain.match(line)
               mode = grp.group(1)
               auditcls = grp.group(2)
               addEvent = True
            elif re_empty.match(line):
               grp = re_empty.match(line)
               mode = grp.group(1)
               auditcls = ""
               addEvent = True
            else:
                self.result.warning("Unrecognized line encountered. Line: " + java.lang.Integer.toString(index+1))
                success = False        

            if addEvent:
               if currentObj == "":
                  self.result.warning("Event line encountered with no parent object. Line: " + java.lang.Integer.toString(index+1))
                  success = False
               else:
                  if self.objects[currentObj].flagExists(mode):
                     self.result.warning("Duplicate action entry '" + mode + "' found for object '" + currentObj + "'. Line: " + java.lang.Integer.toString(index+1))
                     success = False
                  else:
                     self.objects[currentObj].action_entries[mode] = auditcls

        return success


    def matchPolicy(self, policy, reconcile):
        match = True
        files = policy.resourceMap.get(PolicyConstants.FILE)
        audits = policy.ruleMap.get(PolicyConstants.AUDIT)

        if files is None or files.size() == 0:
            return True

        testing_keys = files.keySet().toArray()
        for filename in testing_keys:
            fileObj = files.get(filename)

            filematch = True
            self.result.startAction(fileObj)
            if filename in self.objects:
                # The file exists in both, now we have to check the audit
                if fileObj.getAuditID() != None:
                    auditObj = audits.get(fileObj.getAuditID())
                    if not self.objects[filename].matchPolicy(auditObj):
                        filematch = False
                elif len(self.objects[filename].action_entries) > 0:
                    # The file doesn't have an audit ID
                    for e in self.objects[filename].action_entries:
                        self.result.diff("audit", e, self.objects[filename].action_entries[e], None)

                    del self.objects[filename]
                    filematch = False

            elif fileObj.getAuditID() is not None:
                # This file object in the testing policy doesn't exist in the objects file
                auditObj = audits.get(fileObj.getAuditID())
                self.addObject(fileObj, auditObj)
                filematch = False
        
            if filematch is False:
                match = False
                if reconcile:
                    self.result.finishAction(npm_s_policy_reconciled, [])
                else:
                    self.result.finishAction(npm_s_policy_not_compliant, [])
            else:
                self.result.finishAction(npm_s_success, [])

        return match


    def addObject(self, fileObj, auditObj):
        if "permit" in auditObj.getLevels() or "deny" in auditObj.getLevels():
            objectName = fileObj.getKey()
            if objectName in self.objects:
                self.objects[objectName].matchPolicy(auditObj)
            else:
                self.objects[objectName] = AIXAuditObject(objectName, self.result, True)
                self.objects[objectName].arrPos = len(self.objectsList)
                self.objectsList.append(self.objects[objectName])
                self.objects[objectName].buildEntries(auditObj)


    def getFile(self):
        output = java.lang.StringBuffer("")
        for i in range(len(self.objectsList)):
            if(not self.objectsList[i].inPolicy):
                d = self.objectsList[i].getData()
                if d != "":
                    output.append(d)
                    output.append("\n")

        header = False
        for i in range(len(self.objectsList)):
            if(self.objectsList[i].inPolicy):
                if not header:
                    header = True
                    output.append("*\n* TAMOS Agent is managing entries below this line.\n*\n\n")
                
                d = self.objectsList[i].getData()
                if d != "":
                    output.append(d)
                    output.append("\n")

        return output.toString()

class AIXAuditObject:
    """
    Represents a single entry into the 'objects' file.
    """

    #
    # Globals (created mostly in the constructor)
    #
    # result            - A pointer to the PluginResult object.
    #                     This is used for logging and whatnot.
    # objectName        - The name of the object (the file path).
    # action_entries    - A HashMap of action entries for the object 
    #                     (i.e. r = "FILE_Open")
    #

    def __init__(self, objectName, result, inPolicy = False):
        self.objectName = objectName
        self.result = result
        self.inPolicy = inPolicy
        self.action_entries = {}

    def buildEntries(self, auditObj):
        new_entries = getActionEntries(auditObj)
        for e in new_entries:
            if e not in self.action_entries:
                self.result.diff("audit", e, None, new_entries[e])
            elif self.action_entries[e] != new_entries[e]:
                self.result.diff("audit", e, self.action_entries[e], new_entries[e])

        for e in self.action_entries:
            if e not in new_entries:
                self.result.diff("audit", e, self.action_entries[e], None)

        self.action_entries = getActionEntries(auditObj)

    """
    This function updates the object so that it reflects the policy embodied
    in the auditObj object. It returns 'True' if no changes were needed to
    reflect the policy, ie. it already matches, or 'False' if changes were
    needed to reflect the policy, ie. it didn't already match.
    """
    def matchPolicy(self, auditObj):
        match = True

        self.inPolicy = True
        if "permit" in auditObj.getLevels() or "deny" in auditObj.getLevels():
            action_entries_t = getActionEntries(auditObj)
            existing_keys = self.action_entries.keys()
            testing_keys = action_entries_t.keys()
            for key in testing_keys:
                if key in existing_keys:
                    existing_keys.remove(key)
                    if self.action_entries[key] != action_entries_t[key]:
                        # Action exists in both, event is different
                        self.result.diff("audit", key, self.action_entries[key], action_entries_t[key])
                        self.action_entries[key] = action_entries_t[key]
                        match = False
                else:
                    # Key exists in the test policy but not in existing policy
                    self.result.diff("audit", key, None, action_entries_t[key])
                    self.action_entries[key] = action_entries_t[key]
                    match = False

            # There are action entries in existing policy that aren't in the testing policy
            for key in existing_keys:
                self.result.diff("audit", key, self.action_entries[key], None)
                del self.action_entries[key]
                match = False

        elif len(self.action_entries) > 0:
            # There are action entries in current policy but no audit level is set
            for entry in self.action_entries:
                self.result.diff("audit", entry, self.action_entries[entry], None)

            self.action_entries = {}
            match = False

        return match


    def flagExists(self, flag, value = None):
        if value and flag in self.action_entries and self.action_entries[flag] == value:
            return True
        elif not value and flag in self.action_entries:
            return True
        else:
            return False

    def getData(self):
        if len(self.action_entries) == 0:
            return ""

        d = "%s:\n" % self.objectName
        for action in self.action_entries:
            d = d + "\t%s = \"%s\"\n" % (action, self.action_entries[action])

        return d

# vim: set ts=4 sw=4 expandtab :
