#
# IBM Confidential
# OCO Source Materials
# Tivoli Access Manager for Operating Systems - Agent
# TIVO-TAMOSA
#
# (C) Copyright IBM Corp. 2008
# The source code for this program is not published or otherwise
# divested of its trade secrets, irrespective of what has been
# deposited with the U.S. Copyright Office.
#

#
# solaris file audit implementation
#

from ossutil import wrapPyException
from ossutil.solarisaudit import *
from ossutil.configeditor import LineEditor
from ossutil.tam import parseTAMActionString

from com.ibm.amos.agent.db import PolicyConstants, PolicyDelta, RuleQuery
from com.ibm.amos.agent.db.resources import File
from com.ibm.amos.nls.pdopmidx import *

import java.lang.String
import java.lang.StringBuffer
import java.lang.Exception

# mapping of TAMOS action bits to solaris audit classes
tamos_action_to_audit_class = {     # TAMOS action -> class (syscalls)
    'r': ('fr','fa'),               # read -> file read, file attribute read (open, openat, readlink)
    'w': ('fw',),                   # write -> file write (open, openat)
    'N': ('fc',),                   # create -> file create (creat, link, mknod, symlink, rename, renameat, mkdir, open, openat, core dump (!))
    'x': ('ex',),                   # execute -> execute (exec, execve)
    'o': ('fm',),                   # chown -> file modify (chown, fchown, fchownat)
    'p': ('fm',),                   # chmod -> file modify (chmod, fchmod, acl, facl)
    'D': ('pm',),                   # chdir -> process modify (chdir, chroot, fchdir, fchroot)
    'd': ('fd',),                   # delete -> file delete (rename, renameat, unlink, unlinkat, rmdir, open (trunc))
    'U': ('fm',),                   # utime -> file modify (utimes, utime, futimesat)
    'l': ('fr',)                    # list -> file read (open, openat)
}

tamos_action_to_audit_class_option = {
    'R': ('fc', 'fd')               # rename -> file create or file delete (rename, renameat)
}

def combineAuditClasses(classes, moreclasses):
    for c in moreclasses:
        if c not in classes:
            classes.append(c)

def combineAuditClassOptions(classes, options):
    # if any of the options are already set, do nothing
    for opt in options:
        if opt in classes:
            return
        
    # otherwise add the first one
    classes.append(options[0])

# combine all classes to see what we want to do when no permissions are specified
all_classes = []
for ac in tamos_action_to_audit_class.itervalues():
    combineAuditClasses(all_classes, ac)
for ac in tamos_action_to_audit_class_option.itervalues():
    combineAuditClassOptions(all_classes, ac)

all_classes.sort()


class SolarisAuditControl:
    """
    Solaris audit_control file representation.
    """
    def __init__(self, data):
        self.edit = LineEditor(data)

    def getAuditFlags(self):
        """
        Extracts the current audit flags from the file
        """
        for l in self.edit.lines():
            if l.startswith("flags:"):
                flags = l[len("flags:"):]
                return flags.strip()
        else:
            return ""

    def setAuditFlags(self, flags):
        """
        Updates the audit flags line in the file, creating one if
        there isn't one already
        """
        def cb(line, newflags):
            if line.startswith("flags:"):
                return newflags
            else:
                return line

        newflags = "flags: " + flags
        self.edit.changed = False
        self.edit.update(cb, newflags)

        if self.edit.changed is False:
            # didn't find an existing flags line, so add one
            self.edit.append(newflags)

    def toAuditFile(self):
        """
        Returns the updated contents of the audit_control file
        """
        return self.edit.getData()



class SolarisFileAuditPlugin:
    def __init__(self, realplugin, target):
        self.realplugin = realplugin
        self.target = target
        # probably check bsm is enabled?

    def shutdown(self):
        # do nothing
        pass


    def determineFileAuditPolicy(self, result, policy):
        """
        Figures out which file-related event classes to enable based on the given policy.
        Returns a tuple containing lists of classes to enable for both (0), permit only (1),
        deny only (2) and none (3) auditing.
        """

        rq = RuleQuery(PolicyConstants.AUDIT)
        rq.resourceRuleQuery(PolicyDelta.ALL, PolicyConstants.FILE)
        rules = rq.evaluate(policy)

        # accumulate classes
        both_classes = []
        permit_classes = []
        deny_classes = []
        none_classes = []
        permit_classes_options = []
        deny_classes_options = []
        for audit in rules:

            def mapActionStringFromAttr(audit, attr, classes, class_options):
                actions = audit.getExtAttributeValue(attr, 0)
                if actions is None:
                    # no ext attr -> all mapped classes
                    for a in all_classes:
                        if a not in classes:
                            classes.append(a)
                else:
                    # add the actions specified in the ext attr
                    # find the OSSEAL action group:
                    parsed = parseTAMActionString(actions)
                    osseal = ""
                    if parsed is None:
                        result.actionWarning("Unable to parse action bit string '%s' in attribute '%s' on audit object '%s'" % (actions, attr, audit.getKey()))
                    elif "OSSEAL" in parsed:
                        osseal = parsed["OSSEAL"]

                    for a in osseal:
                        if a in tamos_action_to_audit_class:
                            combineAuditClasses(classes, tamos_action_to_audit_class[a])
                        elif a in tamos_action_to_audit_class_option:
                            # delay handling these until we've got everything else
                            class_options.append(tamos_action_to_audit_class_option[a])
                        else:
                            result.actionWarning("Unknown OSSEAL action bit '%s' encountered in attribute '%s' on audit object '%s'" % (a, attr, audit.getKey()))

            # figure out which classes to enable, based on the audit levels and extended attributes
            # The extended attributes specify distinct sets of actions for permit and deny auditing,
            # which (I presume) only take effect when the audit level itself does.
            result.addRule(audit)
            if "permit" in audit.getLevels():
                mapActionStringFromAttr(audit, PERMIT_ACTIONS, permit_classes, permit_classes_options)
                                
            if "deny" in audit.getLevels():
                mapActionStringFromAttr(audit, DENY_ACTIONS, deny_classes, deny_classes_options)

        # handle any multiple-option bits
        for pco in permit_classes_options:
            combineAuditClassOptions(permit_classes, pco)
        for dco in deny_classes_options:
            combineAuditClassOptions(deny_classes, dco)

        # figure out which classes are enabled for both permit and deny
        for klass in permit_classes:
            if klass in deny_classes:
                both_classes.append(klass)
                        
        # remove them from the permit-only and deny-only class lists
        for klass in both_classes:
            permit_classes.remove(klass)
            deny_classes.remove(klass)
                        
        # figure out which classes to disable altogether
        for klass in all_classes:
            if klass not in both_classes and klass not in permit_classes and klass not in deny_classes:
                none_classes.append(klass)
                        
        # sort everything for consistency
        both_classes.sort()
        permit_classes.sort()
        deny_classes.sort()
        none_classes.sort()
        
        return (both_classes, permit_classes, deny_classes, none_classes)



    def updateAuditFlags(self, result, policy, flags):
        """
        Updates the global audit flags.
        """

        # figure out which classes the policy says should be enabled and disabled
        result.newStep("Determining file audit policy")
        (both, permit, deny, none) = self.determineFileAuditPolicy(result, policy)
        # probably trace the audit class sets

        # append explicit flags to ensure the classes are enabled or disabled
        # in line with the policy
        addflags = ""
        if flags != "":
            addflags = ","
        for klass in both:
            addflags += ",%s" % klass
        for klass in permit:
            addflags += ",+%s,^-%s" % (klass, klass)
        for klass in deny:
            addflags += ",-%s,^+%s" % (klass, klass)
        for klass in none:
            addflags += ",^%s" % klass

        if len(addflags) > 1:
            flags += addflags

        # simplify the flags
        norm = normalizeAuditFlags(flags, audit_metaclass_map)
        return norm



    def applyUpdatedFlags(self, result, audit_control, flags):
        """
        Updates and reloads the audit configuration on the target system.
        """

        result.newStep("Updating audit configuration file")
        audit_control.setAuditFlags(flags)

        xac = self.target.translatePath(AUDIT_CONTROL)
        self.target.putFile(xac, audit_control.toAuditFile(), True)

        result.newStep("Reloading audit configuration")
        out = java.lang.StringBuffer()
        err = java.lang.StringBuffer()
        rc = self.target.execute(("/usr/sbin/audit", "-s"), None, out, err)        # FIXME permissions
        if rc != 0:
            # NEWMSG
            result.finishAction(-1, "Reloading audit configuration failed: %s" % rc)
            return -1
        else:
            return npm_s_success



    def policyUpdate(self, result, delta, apply):
 
        outcome = npm_s_success
        try:
            # represent this as a single action on the audit_control file,
            # even though that probably doesn't exist in the policy
            result.startAction(File(AUDIT_CONTROL, None, None))

            # should parse audit_classes, audit_events here
            # to make sure the classes are as we expect
            
            xac = self.target.translatePath(AUDIT_CONTROL)

            # read the audit_control file, extract the current flags
            # tiny permission problem here at present..
            data = java.lang.String(self.target.getFile(xac))
            audit_control = SolarisAuditControl(data)
            oldflags = audit_control.getAuditFlags()

            # calculate the new flags line
            flags = self.updateAuditFlags(result, delta.newPolicy, oldflags)

            # if it's materially different to the previous one, update it
            if compareFlags(flags, oldflags, audit_metaclass_map) == False:
                result.diff(AUDIT_CONTROL, None, oldflags, flags)
                if apply:
                    outcome = self.applyUpdatedFlags(result, audit_control, flags)
                else:
                    outcome = npm_s_policy_not_compliant
            else:
                # probably note that nothing needed to be done?
                pass

            result.finishAction(outcome, None)
            result.finish(outcome, None)
        except java.lang.Exception, e:
            result.finish(e)
        except Exception, e:
            result.finish(wrapPyException(self.realplugin, e))
 

    def verifyPolicy(self, result, policy, reconcile):

        try:
            outcome = npm_s_success

            # represent this as a single action on the audit_control file,
            # even though that probably doesn't exist in the policy
            result.startAction(File(AUDIT_CONTROL, None, None))

            # figure out policy
            (both, permit, deny, none) = self.determineFileAuditPolicy(result, policy)
                    
            # get current audit flags
            xac = self.target.translatePath(AUDIT_CONTROL)
            data = java.lang.String(self.target.getFile(xac))
            audit_control = SolarisAuditControl(data)

            flags = audit_control.getAuditFlags()
                    
            compliant = True
            for klass in both:
                if evaluateAuditClass(flags, klass, audit_metaclass_map) != (True, True):
                    compliant = False

            for klass in permit:
                if evaluateAuditClass(flags, klass, audit_metaclass_map) != (True, False):
                    compliant = False

            for klass in deny:
                if evaluateAuditClass(flags, klass, audit_metaclass_map) != (False, True):
                    compliant = False

            for klass in none:
                if evaluateAuditClass(flags, klass, audit_metaclass_map) != (False, False):
                    compliant = False

            if not compliant:
                newflags = self.updateAuditFlags(result, policy, flags)
                result.diff(AUDIT_CONTROL, None, flags, newflags)
                if reconcile:
                    result.newStep("Reconciling system audit configuration with file audit policy")
                    outcome = self.applyUpdatedFlags(result, audit_control, newflags)
                    if outcome == npm_s_success:
                        outcome = npm_s_policy_reconciled
                else:
                    outcome = npm_s_policy_not_compliant

            result.finishAction(outcome, [])


            result.finish(outcome, [])
        except java.lang.Exception, e:
            result.finish(e)
        except Exception, e:
            result.finish(wrapPyException(self.realplugin, e))


# vim: set ts=4 sw=4 expandtab :
