#
# IBM Confidential
# OCO Source Materials
# Tivoli Access Manager for Operating Systems - Agent
# TIVO-TAMOSA
#
# (C) Copyright IBM Corp. 2008
# The source code for this program is not published or otherwise
# divested of its trade secrets, irrespective of what has been
# deposited with the U.S. Copyright Office.
#

"""
    ACL format support for AIX Classic ACLs
"""

import fsacl.convert
from ossutil.policydiff import compareStringMap

from com.ibm.amos.agent.db import PolicyConstants
from com.ibm.amos.agent.db.rules import ACL, ACLEntry
from com.ibm.amos.nls.pdopmidx import *

from java.lang import IllegalStateException
from java.io import IOException

# TAMOS mapping bits (unix bit -> set of TAMOS bits)
# The mappings here are the same as for the POSIX draft ACLs
tamosFileBits = {
    'r': 'r',
    'w': 'w',
    'x': 'x'
}

tamosDirBits = {
    'r': 'l',
    'w': 'NRUdop',
    'x': 'TD'
}

__aixctypes = (
    "specify",
    "permit",
    "deny"
)

__aixcperms = (
    'r', 'w', 'x'
)

def convertToArgs(result, acl):
    """
        Converts an ACL in AIXC format to a sequence of arguments for the
        pdosfsacl tool
    """

    args = ["aixc"]

    # unix entries
    if (acl.getOwnerPermission() is not None):
        args += [ "unix:" + acl.getOwnerPermission() + ":user" ]

    if (acl.getOwningGroupPermission() is not None):
        args += [ "unix:" + acl.getOwningGroupPermission() + ":group" ]

    if (acl.getAnyOtherPermission() is not None):
        args += [ "unix:" + acl.getAnyOtherPermission() + ":other" ]
    else:
        args += [ "unix::other" ]
    
    # user entries, sort them for predictability
    users = list(acl.userMap.keySet())
    users.sort()
    for u in users:
        args += [ "specify:" + acl.getUserPermission(u) + ":u:" + u ]

    # group entries, also sorted
    groups = list(acl.groupMap.keySet())
    groups.sort()
    for g in groups:
        args += [ "specify:" + acl.getGroupPermission(g) + ":g:" + g ]

    return args

def addUnixACLEntry(acl, type, perms, subject):
    
    entry = ACLEntry()
    entry.permission = perms
    if subject == "user":
        entry.entryClass = ACLEntry.OWNER_ACL_ENTRY
    elif subject == "group":
        entry.entryClass = ACLEntry.OWNING_GROUP_ACL_ENTRY
    elif subject == "other":
        entry.entryClass = ACLEntry.ANY_OTHER_ACL_ENTRY
    else:
        raise IllegalStateException("Invalid subject for unix entry: " + subject)
    
    acl.addACLEntry(entry)

def addAIXCACLEntry(acl, type, perms, reqliststr):
    """
        Converts AIXC ACL entry output from pdosfsacl tool to an ACL in
        AIXC format - currently TAMOSH only supports setting 'specify'
        entries that contain a single user / group entry, so any
        'permit' or 'deny' entries, or entries that contain more than
        one entity in the reqliststr are mapped to the unsupported type.
    """

    entry = ACLEntry()
    entry.permission = perms

    reqlist = reqliststr.split(",")

    if len(reqlist) > 1:
        acl.addExtAttribute("unsupported", type + ":" + perms + ":" + reqliststr)
    else:
        try:
            (subjectType, subject) = reqliststr.split(":")
        except ValueError, e:
            raise IllegalStateException("Unable to parse output line from ACL tool: " + reqliststr)

        if not subjectType:
            raise IllegalStateException("Unable to determine subject type from ACL tool: " + reqliststr)
        elif not subject:
            raise IllegalStateException("Unable to determine subject from ACL tool: " + reqliststr)

        entry = ACLEntry()
        entry.permission = perms
        
        if type == "specify":
            entry.subject = subject
            if subjectType == "u":
                entry.entryClass = ACLEntry.USER_ACL_ENTRY
            elif subjectType == "g":
                entry.entryClass = ACLEntry.GROUP_ACL_ENTRY
            else:
                raise IllegalStateException("Invalid subject type for specify entry: " + subjectType)

            # We have a valid ACL entry
            acl.addACLEntry(entry)

        else:
            acl.addExtAttribute("unsupported", type + ":" + perms + ":" + subjectType + ":" + subject)

    return entry
    
        
def convertToACL(result, acl, input):
    """
        Converts output from the pdosfsacl tool to an ACL object in AIXC
        format
    """

    acl.setSource(PolicyConstants.ACL_TYPE_AIXC)
    while True:
        try:
            entryLine = input.readLine()
        except IOException, e:
            raise IllegalStateException("Broke while reading output from ACL tool: " + str(e))

        if entryLine is None:
            break
        try:
            (type, perms, remainder) = entryLine.split(":", 2)
        except ValueError, e:
            raise IllegalStateException("Unable to parse output from ACL tool: " + entryLine)

        # validate the permissions
        for perm in perms:
            if perm not in __aixcperms:
                raise IllegalStateException("Invalid permissions for ACL entry: " + perms)

        # convert / add the acl entry
        if type == "unix":
            addUnixACLEntry(acl, type, perms, remainder)
        elif type in __aixctypes:
            addAIXCACLEntry(acl, type, perms, remainder)
        else:
            raise IllegalStateException("Invalid type for ACL entry: " + type)
            

def compareACLs(result, a, b):
    """
    Returns True if ACLs A and B match in all important respects for the
    format
    """
    r = True
    if a.getOwnerPermission() != b.getOwnerPermission():
        result.diff("owner", None, a.getOwnerPermission(), b.getOwnerPermission())
        r = False

    if a.getOwningGroupPermission() != b.getOwningGroupPermission():
        result.diff("owning-group", None, a.getOwningGroupPermission(), b.getOwningGroupPermission())
        r = False
        
    if a.getAnyOtherPermission() != b.getAnyOtherPermission():
        result.diff("any-other", None, a.getAnyOtherPermission(), b.getAnyOtherPermission())
        r = False

    if a.getExtAttribute("unsupported") != b.getExtAttribute("unsupported"):
        result.diff("unsupported", None, a.getExtAttribute("unsupported"), b.getExtAttribute("unsupported"))
        r = False

    if compareStringMap(result, "user", a.userMap, b.userMap) == False:
        r = False

    if compareStringMap(result, "group", a.groupMap, b.groupMap) == False:
        r = False

    return r


def mapACLBits(map, perms):
    """
        Maps a TAMOS permission string to a Unix-style rwx string according
        to a specific map of output bits to a string of input bits.
        This mapping is the same as the POSIX ACL mapping.
    """

    if perms is None:
        return None

    result = ""

    # since we're in POSIX-specific code here, we only handle r w and x, in that order.
    for p in "rwx":
        if p not in map:
            continue
        
        for b in map[p]:
            if perms.find(b) != -1:
                result += p 
                break
    return result

def convertFromTAMOS(result, currentFile, currentAcl, policyFile, policyAcl):
    """
        Converts a TAMOS-format ACL to an AIXC ACL as best (or easily) we can,
        taking the file type and current permissions into account where
        appropriate
    """

    if currentFile.getExtAttributeValue(PolicyConstants.UNIX_FILE_TYPE, 0) == \
       PolicyConstants.FILE_TYPE_DIRECTORY:
        bitmap = tamosDirBits
    else:
        bitmap = tamosFileBits

    return fsacl.convert.convertFromTAMOS(PolicyConstants.ACL_TYPE_AIXC, result, currentFile, currentAcl, policyFile, policyAcl, mapACLBits, bitmap)

def convertFromNone(result, currentFile, currentAcl, policyFile, policyAcl):
    """
    Constructs an AIXC ACL that contains only the basic unix permissions.
    This is used when there is no ACL attached to the file object, but we still
    need to apply policy to the file.
    """
    acl = ACL()
    acl.setSource(PolicyConstants.ACL_TYPE_AIXC)

    # copy owner, owning-group, and any-other permissions from the current ACL
    acl.setOwnerPermission(currentAcl.getOwnerPermission())
    acl.setOwningGroupPermission(currentAcl.getOwningGroupPermission())
    acl.setAnyOtherPermission(currentAcl.getAnyOtherPermission())
    return acl

# vim: set ts=4 sw=4 expandtab :
