"""
ACL format support for NFSv4 ACLs
"""

import fsacl.convert
from ossutil.policydiff import compareStringMap

from com.ibm.amos.agent.db.rules import ACL, ACLEntry
from com.ibm.amos.agent.db import PolicyConstants
from com.ibm.amos.nls.pdopmidx import *

from java.lang import IllegalStateException

# attribute used to mark ACLs that were not fully converted
CONVERSION_FAILED = "nfsv4-conversion-failed"
EXTRANEOUS_FOUND = "nfsv4-extraneous-found"
EXISTING_OWNER_ALLOW = "nfsv4-owner-allow"
EXISTING_OWNER_DENY = "nfsv4-owner-deny"
EXISTING_OWNING_GROUP_ALLOW = "nfsv4-owning-group-allow"
EXISTING_OWNING_GROUP_DENY = "nfsv4-owning-group-deny"

ORIGINAL_ENTRIES = "nfsv4-original-entries"

# NFSv4 permissions implicitly granted to the resource owner
IMPLIED_OWNER_PERMS = "WcCA"

# NFSv4 permissions in order
nfsv4Perms = ('r', 'w', 'p', 'R', 'A', 'x', 'a', 'W', 'd', 'D', 'c', 'C', 'o', 's')

# mapping from TAMOS action bits to lists of NFSv4 permissions

# for files:
tamosFileBits = {
    ''  : ('R', 'a', 'c', 's'),
    'N' : ('w',),      # does this make any sense for files?
    'R' : ('w','p','d'),
    'U' : ('A',),
    'd' : ('d',),
    'o' : ('o',),
    'p' : ('A','C'),
    'r' : ('r',),
    'w' : ('w','p','W'),
    'x' : ('x',)
}

# and for directories:
tamosDirBits = {
    ''  : ('R', 'a', 'c', 's'),
    'D' : ('x',),
    'N' : ('w','p','W'),
    'R' : ('w','p','d','D'),
    'U' : ('A',),
    'd' : ('d','D'),
    'l' : ('r'),
    'o' : ('o',),
    'p' : ('A','C'),
    'T' : ('x',)
}

def invertNFSv4Perms(perms):
    inv = ""
    for p in nfsv4Perms:
        if perms is None or p not in perms:
            inv += p
    return inv

def combineNFSv4Perms(strings):
    comb = ""
    for perm in nfsv4Perms:
        for s in strings:
            if s and perm in s:
                comb += perm
                break
    return comb

def subtractNFSv4Perms(a, b):
    r = ""
    for perm in nfsv4Perms:
        if perm in a and perm not in b:
            r += perm
    return r

def mapACLBits(bitmap, permissions):
    if permissions is None:
        return None

    mapped = {}
    if '' in bitmap:
        for b in bitmap['']:
            mapped[b] = 1

    for p in permissions:
        if p not in bitmap:
            continue

        for b in bitmap[p]:
            mapped[b] = 1

    s = ""
    for b in nfsv4Perms:
        if b in mapped:
            s += b

    return s

  
# TAMOS user and group entries are represented using pairs of
# NFSv4 entries, allowing the actions specified in the ACL
# and denying all others, to ensure that only the most specific
# entry in the ACL applies.  When converting an NFSv4 format ACL
# back to TAMOS format, we need to match the allow and deny entries up.

def combineEntries(current, sense, perms):
    if sense == 'allow':
        if current is not None:
            # multiple allow entries -> drop all permissions
            return None
        
        # prefix the entry to mark that we haven't seen the deny NFSv4 entry yet
        return "!" + combineNFSv4Perms((perms,))

    elif sense == 'deny':
        if current is None or current[0] != '!':
            # haven't seen the allow entry yet, or we've already seen a deny entry -> drop it
            return None
        
        # deny entry actions must be the inverse of the allow entry
        if invertNFSv4Perms(perms) != current[1:]:
            # deny entry doesn't match -> drop it
            return None

        # OK, we've got the allow and and the complementary deny
        return current[1:]  # remove the '!' prefix
    else:
        # unexpected entry type.. ignore it?
        return current


# things that convert between NFSv4 ACL entries and TAM-like entries.
# we don't currently attempt to fully represent NFSv4 ACLs in our ACL
# object, only the subset that look like ACLs we would create from TAMOS
# ACLs.

class SingleInstanceEntry:
    """
    Base class for entry types that can only appear once.
    (owner, owning-group)
    """
    def addToACL(self, acl, sense, type, subject, perms):
        if type != self.entryType():
            return False

        newPerms = combineEntries(self.get(acl), sense, perms)
        self.set(acl, newPerms)
        return True

    def convertToArgs(self, acl, args):
        perms = self.get(acl)
        args += (self.entryType(), "", perms or "", "allow")
        args += (self.entryType(), "", invertNFSv4Perms(perms), "deny")

    def complete(self, acl):
        perms = self.get(acl)
        if perms is not None and perms != "" and perms[0] == '!':
            self.set(acl, None)
            return False

        return True


class OwnerEntry(SingleInstanceEntry):
    def entryType(self):
        return ACLEntry.OWNER_ACL_ENTRY

    def convertToArgs(self, acl, args):
        perms = self.get(acl)
        args += (self.entryType(), "", perms or "", "allow")
        # exclude permissions implicitly granted to owner; on AIX
        # systems, attempting to deny these causes the ACL application
        # to fail.  Solaris just disregards the ACL for these operations.
        deny = subtractNFSv4Perms(invertNFSv4Perms(perms), IMPLIED_OWNER_PERMS)
        args += (self.entryType(), "", deny, "deny")

    def get(self, acl):
        return acl.getOwnerPermission()
    
    def set(self, acl, perms):
        acl.setOwnerPermission(perms)


class OwningGroupEntry(SingleInstanceEntry):
    def entryType(self):
        return ACLEntry.OWNING_GROUP_ACL_ENTRY

    def get(self, acl):
        return acl.getOwningGroupPermission()
    
    def set(self, acl, perms):
        acl.setOwningGroupPermission(perms)


class MultiInstanceEntry:
    """
    Base class for entry types that can appear any number of times
    (user, group)
    """
    def addToACL(self, acl, sense, type, subject, perms):
        if type != self.entryType():
            return False

        newPerms = combineEntries(self.get(acl, subject), sense, perms)
        self.set(acl, subject, newPerms)
        return True
  
    def convertToArgs(self, acl, args):
        subjects = self.getSubjects(acl)
        subjects.sort()
        for s in subjects:
            perms = self.get(acl, s)
            args += (self.entryType(), s, perms, "allow")
  
        for s in subjects:
            perms = self.get(acl, s)
            args += (self.entryType(), s, invertNFSv4Perms(perms), "deny")
  
    def complete(self, acl):
        result = True
        subjects = self.getSubjects(acl)
        for s in subjects:
            perms = self.get(acl, s)
            if perms is not None and perms != "" and perms[0] == '!':
                self.remove(acl, s)
                result = False
  
        return result
  
  
class UserEntry(MultiInstanceEntry):
    def entryType(self):
        return ACLEntry.USER_ACL_ENTRY

    def get(self, acl, subject):
        return acl.getUserPermission(subject)

    def set(self, acl, subject, perms):
        acl.addUserPermission(subject, perms)

    def remove(self, acl, subject):
        # ugly
        acl.userMap.remove(subject)

    def getSubjects(self, acl):
        return list(acl.userMap.keySet())
        


class GroupEntry(MultiInstanceEntry):
    def entryType(self):
        return ACLEntry.GROUP_ACL_ENTRY

    def get(self, acl, subject):
        return acl.getGroupPermission(subject)

    def set(self, acl, subject, perms):
        acl.addGroupPermission(subject, perms)

    def remove(self, acl, subject):
        # ugly
        acl.groupMap.remove(subject)

    def getSubjects(self, acl):
        return list(acl.groupMap.keySet())
        

class EveryoneEntry:
    def entryType(self):
        return ACLEntry.EVERYONE_ACL_ENTRY

    def addToACL(self, acl, sense, type, subject, perms):
        if type != ACLEntry.EVERYONE_ACL_ENTRY:
            return False

        # no allow+deny combination here, just take any allowed permissions
        if acl.getAnyOtherPermission() is not None:
            return False

        if sense == 'allow':
            acl.setAnyOtherPermission(perms)
        return True

    def convertToArgs(self, acl, args):
        everyone = acl.getAnyOtherPermission()
        if everyone is not None:
            args += (ACLEntry.EVERYONE_ACL_ENTRY, "", everyone, "allow")

    def complete(self, acl):
        if acl.getAnyOtherPermission() is None:
            acl.setAnyOtherPermission("")
        return True

class ExtraneousEntry:
    """
    Soaks up an extra entries, doesn't add them to the ACL, and ensures we
    treat the ACL as noncompliant.
    """

    def entryType(self):
        return ""

    def addToACL(self, acl, sense, type, subject, perms):
        acl.addExtAttribute(EXTRANEOUS_FOUND, "true")
        return True

    def convertToArgs(self, acl, args):
        # do nothing
        pass

    def complete(self, acl):
        return acl.getExtAttributeValue(EXTRANEOUS_FOUND, 0) is None


# we create entries in a specific order (by entry type) that results in 
# the ACLs evaluating (hopefully) the same as TAM ACLs do.  when reading
# an ACL from the filesystem, we expect to find the entries in the same order,
# giving up if they're not.
entry_order = (
    OwnerEntry(),
    UserEntry(),
    OwningGroupEntry(),
    GroupEntry(),
    EveryoneEntry(),
    ExtraneousEntry()
)


def convertToArgs(result, acl):
    args = ["nfsv4"]

    for e in entry_order:
        e.convertToArgs(acl, args)

    return args



def convertToACL(result, acl, input):
    """
    Converts output from the pdosfsacl tool to an ACL object in NFSv4 format
    """
    acl.setSource(PolicyConstants.ACL_TYPE_NFSV4)
    epi = iter(entry_order)
    entry = epi.next()
    failed = False

    # the theory here is that we only need to recognise ACLs that we would have created
    # ourselves - anything else is not going to match the policy, so we'll either replace
    # it (if applying or reconciling) or report it as a different (if verifying)
    while True:
        # get next line
        try:
            entryLine = input.readLine()
        except IOException, e:
            raise IllegalStateException("Broke while reading output from ACL tool: " + str(e))

        if entryLine is None:
            break

        # remember the ACL as we received it
        acl.addExtAttribute(ORIGINAL_ENTRIES, entryLine)

        # split it into the various fields we use
        try:
            (sense, type, subject, perms, inherit) = entryLine.split(":")
        except ValueError, e:
            raise IllegalStateException("Unable to parse output line from ACL tool: " + entryLine)

        # take particular note of owner@ and group@ entries.  since we try to preserve those
        # if the policy doesn't provide any permissions, we need to remember them even if they
        # don't look like entries we would have created.  ignore inherit-only entries, as we
        # ignore inheritance entirely.
        if "i" not in inherit:
            attrs = []
            if type == ACLEntry.OWNER_ACL_ENTRY or type == ACLEntry.EVERYONE_ACL_ENTRY:
                attrs.append([EXISTING_OWNER_ALLOW, EXISTING_OWNER_DENY])

            if type == ACLEntry.OWNING_GROUP_ACL_ENTRY or type == ACLEntry.EVERYONE_ACL_ENTRY:
                attrs.append([EXISTING_OWNING_GROUP_ALLOW, EXISTING_OWNING_GROUP_DENY])

            for (allow, deny) in attrs:
                pallow = acl.getExtAttributeValue(allow, 0)
                pdeny = acl.getExtAttributeValue(deny, 0)
                if sense == 'allow':
                    # add these permissions to those allowed, but only if
                    # they're not already denied
                    if pdeny:
                        perms = subtractNFSv4Perms(perms, pdeny)
                    pallow = combineNFSv4Perms([pallow, perms])
                elif sense == 'deny':
                    # add these permissions to the set already denied
                    pdeny = combineNFSv4Perms([pdeny, perms])
                
                if pallow is not None:
                    acl.removeExtAttribute(allow)
                    acl.addExtAttribute(allow, pallow)
                if pdeny is not None:
                    acl.removeExtAttribute(deny)
                    acl.addExtAttribute(deny, pdeny)

        # we currently don't set inheritance flags on any entries we create, so
        # if the entry has any set, we'll just ignore it
        # except that our deny entries seem to have file ('f') inheritance set.
        if inherit != "" and inherit != "f":
            failed = True
            continue

        # attempt to add it to the ACL; if the current entry handler can't process it,
        # move on to the next.
        try:
            while entry.addToACL(acl, sense, type, subject, perms) == False:
                entry = epi.next()
        except StopIteration:
            # this actually can't happen.
            failed = True
            break

    # make sure all the entry types are happy
    for e in entry_order:
        if e.complete(acl) == False:
            failed = True
           
    if failed:
        acl.addExtAttribute(CONVERSION_FAILED, "true")
    
    # if we didn't find owner@/group@ entries as we would have set them,
    # use whatever permissions were allowed there
    if acl.getOwnerPermission() is None:
        eop = acl.getExtAttributeValue(EXISTING_OWNER_ALLOW, 0) or ""
        acl.setOwnerPermission(eop)

    if acl.getOwningGroupPermission() is None:
        egp = acl.getExtAttributeValue(EXISTING_OWNING_GROUP_ALLOW, 0)
        acl.setOwningGroupPermission(egp)



def compareACLs(result, a, b):
    # first check if either ACL was not fully converted from actual NFSv4 format,
    # in which case we must treat them as being different
    if a.getExtAttributeValue(CONVERSION_FAILED, 0) or \
       b.getExtAttributeValue(CONVERSION_FAILED, 0):
        # no specific changes to describe
        return False

    # otherwise just compare all the details.
    # hmm, this looks just like every other acl compare function..
    r = True
    if a.getOwnerPermission() != b.getOwnerPermission():
        result.diff("owner", None, a.getOwnerPermission(), b.getOwnerPermission())
        r = False
    
    if a.getOwningGroupPermission() != b.getOwningGroupPermission():
        result.diff("owning-group", None, a.getOwningGroupPermission(), b.getOwningGroupPermission())
        r = False
    
    if a.getAnyOtherPermission() != b.getAnyOtherPermission():
        result.diff("any-other", None, a.getAnyOtherPermission(), b.getAnyOtherPermission())
        r = False

    if compareStringMap(result, "user", a.userMap, b.userMap) == False:
        r = False

    if compareStringMap(result, "group", a.groupMap, b.groupMap) == False:
        r = False

    return r


def convertFromTAMOS(result, currentFile, currentAcl, policyFile, policyAcl):
    """
    Converts a TAMOS-format ACL to our internal representation of an NFSv4 ACL.
    """

    if currentFile.getExtAttributeValue(PolicyConstants.UNIX_FILE_TYPE, 0) == PolicyConstants.FILE_TYPE_DIRECTORY:
        bitmap = tamosDirBits
    else:
        bitmap = tamosFileBits

    return fsacl.convert.convertFromTAMOS(PolicyConstants.ACL_TYPE_NFSV4, result, currentFile, currentAcl, policyFile, policyAcl, mapACLBits, bitmap)

def convertFromNone(result, currentFile, currentAcl, policyFile, policyAcl):
    """
    Constructs an NFSv4 ACL that effectively only sets the basic unix permissions.
    """
    acl = ACL()
    acl.setSource(PolicyConstants.ACL_TYPE_NFSV4)

    # copy owner, owning-group, and any-other permissions from the current ACL
    acl.setOwnerPermission(currentAcl.getOwnerPermission())
    acl.setOwningGroupPermission(currentAcl.getOwningGroupPermission())
    acl.setAnyOtherPermission(currentAcl.getAnyOtherPermission())
    return acl

def convertFromAIXC(result, currentFile, currentAcl, policyFile, policyAcl):
    """
    Constructs an NFSv4 ACL that is equivalent to the supplied AIX Classic
    ACL. At this point, only takes note of the basic permissions. 
    """
    acl = ACL()
    acl.setSource(PolicyConstants.ACL_TYPE_NFSV4)

    # These mappings are taken from examining the output of aclconvert
    defaultOwnerPerms = "RWAC"
    aixcFileBits = {
        ''  : "adcs",
        'r' : "rR",
        'w' : "wpW",
        'x' : "x"
    }

    aixcDirBits = {
        ''  : "adcs",
        'r' : "rR",
        'w' : "wpWD",
        'x' : "x"
    }

    if currentFile.getExtAttributeValue(PolicyConstants.UNIX_FILE_TYPE, 0) == PolicyConstants.FILE_TYPE_DIRECTORY:
        bitmap = aixcDirBits
    else:
        bitmap = aixcFileBits

    ownerPerms = mapACLBits(bitmap, currentAcl.getOwnerPermission())
    ownerPerms = combineNFSv4Perms([ownerPerms, defaultOwnerPerms])
    acl.setOwnerPermission(ownerPerms)
    acl.setOwningGroupPermission(mapACLBits(bitmap, currentAcl.getOwningGroupPermission()))
    acl.setAnyOtherPermission(mapACLBits(bitmap, currentAcl.getAnyOtherPermission()))

    return acl
    
    
# vim: set ts=4 sw=4 expandtab :

