#
# IBM Confidential
# OCO Source Materials
# Tivoli Access Manager for Operating Systems - Agent
# TIVO-TAMOSA
#
# (C) Copyright IBM Corp. 2008
# The source code for this program is not published or otherwise
# divested of its trade secrets, irrespective of what has been
# deposited with the U.S. Copyright Office.
#


"""
ACL format support for POSIX-draft ACLs
"""

import fsacl.convert
from ossutil.policydiff import compareStringMap

from com.ibm.amos.agent.db import PolicyConstants
from com.ibm.amos.agent.db.rules import ACL, ACLEntry
from com.ibm.amos.nls.pdopmidx import *

from java.lang import IllegalStateException
from java.io import IOException

__supported = ( 
    ACLEntry.USER_ACL_ENTRY, 
    ACLEntry.GROUP_ACL_ENTRY, 
    ACLEntry.OWNER_ACL_ENTRY, 
    ACLEntry.OWNING_GROUP_ACL_ENTRY, 
    ACLEntry.ANY_OTHER_ACL_ENTRY
)
__ignored = (
    "mask"
)

# TAMOS mapping bits (unix bit -> set of TAMOS bits)
tamosFileBits = {
    'r': 'r',
    'w': 'w',
    'x': 'x'
}
tamosDirBits = {
    'r': 'l',
    'w': 'NRUdop',
    'x': 'TD'
}

def convertToArgs(result, acl):
    """ 
    Converts an ACL in POSIX format to a sequence of arguments for the
    pdosfsacl tool.
    """
    
    args = ["posix"]
    
    # basic entries
    args += (ACLEntry.OWNER_ACL_ENTRY, "", acl.getOwnerPermission() or "")
    args += (ACLEntry.OWNING_GROUP_ACL_ENTRY, "", acl.getOwningGroupPermission() or "")
    args += (ACLEntry.ANY_OTHER_ACL_ENTRY, "", acl.getAnyOtherPermission() or "")
    
    # user entries; sort them for predictability
    users = list(acl.userMap.keySet())
    users.sort()
    for u in users:
        args += (ACLEntry.USER_ACL_ENTRY, u, acl.getUserPermission(u))
    
    # group entries, also sorted
    groups = list(acl.groupMap.keySet())
    groups.sort()
    for g in groups:
        args += (ACLEntry.GROUP_ACL_ENTRY, g, acl.getGroupPermission(g))

    # we don't add the mask entry here, the pdosfsacl tool calculates it based on the
    # entries it actually applies.
    return args


def convertToACL(result, acl, input):
    """
    Converts output from the pdosfsacl tool to an ACL object in POSIX format
    """
    acl.setSource(PolicyConstants.ACL_TYPE_POSIX)
    while True:
        try:
            entryLine = input.readLine()
        except IOException, e:
            raise IllegalStateException("Broke while reading output from ACL tool: " + str(e))
        
        if entryLine is None:
            break
        
        try:
            # this fails unless the split() results in exactly three values
            (type, subject, perm) = entryLine.split(":")
        except ValueError, e:
            raise IllegalStateException("Unable to parse output line from ACL tool: " + entryLine)
        
        if type in __supported:
            entry = ACLEntry()
            entry.permission = perm
            entry.subject = subject
            entry.entryClass = type
            acl.addACLEntry(entry)
        elif type not in __ignored:
            raise IllegalStateException("Unsupported ACL entry type: " + type)


def compareACLs(result, a, b):
    """
    Returns True if ACLs A and B match in all important respects for the format
    """
    r = True
    if a.getOwnerPermission() != b.getOwnerPermission():
        result.diff("owner", None, a.getOwnerPermission(), b.getOwnerPermission())
        r = False

    if a.getOwningGroupPermission() != b.getOwningGroupPermission():
        result.diff("owning-group", None, a.getOwningGroupPermission(), b.getOwningGroupPermission())
        r = False
        
    if a.getAnyOtherPermission() != b.getAnyOtherPermission():
        result.diff("any-other", None, a.getAnyOtherPermission(), b.getAnyOtherPermission())
        r = False

    if compareStringMap(result, "user", a.userMap, b.userMap) == False:
        r = False

    if compareStringMap(result, "group", a.groupMap, b.groupMap) == False:
        r = False

    return r


def mapACLBits(map, perms):
    """ 
    Maps a TAMOS permission string to a Unix-style rwx string according to
    a specific map of output bits to a string of input bits.
    """
    if perms is None:
        return None

    result = ""
    
    # since we're in POSIX-specific code here, we only handle r w and x, in that order.
    for p in "rwx":
        if p not in map:
            continue
        
        for b in map[p]:
            if perms.find(b) != -1:
                result += p 
                break
    return result

def convertFromTAMOS(result, currentFile, currentAcl, policyFile, policyAcl):
    """
    Converts a TAMOS-format ACL to a POSIX ACL as best we can, taking the file type and
    current permissions into account where appropriate.
    """

    if currentFile.getExtAttributeValue(PolicyConstants.UNIX_FILE_TYPE, 0) == PolicyConstants.FILE_TYPE_DIRECTORY:
        bitmap = tamosDirBits
    else:
        bitmap = tamosFileBits

    return fsacl.convert.convertFromTAMOS(PolicyConstants.ACL_TYPE_POSIX, result, currentFile, currentAcl, policyFile, policyAcl, mapACLBits, bitmap)

def convertFromNone(result, currentFile, currentAcl, policyFile, policyAcl):
    """
    Constructs a POSIX ACL that contains only the basic unix permissions.
    This is used when there is no ACL attached to the file object, but we still
    need to apply policy to the file.
    """
    acl = ACL()
    acl.setSource(PolicyConstants.ACL_TYPE_POSIX)

    # copy owner, owning-group, and any-other permissions from the current ACL
    acl.setOwnerPermission(currentAcl.getOwnerPermission())
    acl.setOwningGroupPermission(currentAcl.getOwningGroupPermission())
    acl.setAnyOtherPermission(currentAcl.getAnyOtherPermission())
    return acl

# vim: set ts=4 sw=4 expandtab :
