#
# IBM Confidential
# OCO Source Materials
# Tivoli Access Manager for Operating Systems - Agent
# TIVO-TAMOSA
#
# (C) Copyright IBM Corp. 2008
# The source code for this program is not published or otherwise
# divested of its trade secrets, irrespective of what has been
# deposited with the U.S. Copyright Office.
#

# filesystem ACL plugin

from com.ibm.amos.agent.db import PolicyConstants, ResourceQuery, PolicyDelta
from com.ibm.amos.agent.db.rules import ACL
from com.ibm.amos.agent.db.resources import File
from com.ibm.amos.agent.npm import PolicyPlugin, PluginException
from com.ibm.amos.nls.pdopmidx import *
from com.ibm.amos.nls import MessageCatalog

from java.util.logging import Logger, Level
from java.io import LineNumberReader, StringReader
from java.lang import StringBuffer, IllegalStateException
import java.lang.Exception

from ossutil import wrapPyException, splitDirName
from ossutil.configeditor import LineEditor
import ossutil.wildcard

from com.ibm.amos.agent.db.PolicyConstants import RESOURCE_OWNER, RESOURCE_OWNING_GROUP, UNIX_FILE_SPECIAL_BITS

import POSIX
import NoACL
import NFSv4
import AIXC

import re

class FilesystemACLPlugin(PolicyPlugin):
    """
    Native policy manager plugin for controlling filesystem ACLs
    """
    
    def __init__(self):
        # acl formats that can be set, in preference order (earlier ones
        # have higher priority)
        self.__setableFormats = (
            PolicyConstants.ACL_TYPE_NFSV4,
            PolicyConstants.ACL_TYPE_POSIX,
            PolicyConstants.ACL_TYPE_AIXC
        )

        # register ACL formats
        self.__formats = {
            PolicyConstants.ACL_TYPE_POSIX : POSIX,
            PolicyConstants.ACL_TYPE_NFSV4 : NFSv4,
            PolicyConstants.ACL_TYPE_NONE  : NoACL,
            PolicyConstants.ACL_TYPE_AIXC  : AIXC
        }
        
        # and converters
        self.__converters = {
             (PolicyConstants.ACL_TYPE_TAMOS, PolicyConstants.ACL_TYPE_POSIX) : POSIX.convertFromTAMOS,
             (PolicyConstants.ACL_TYPE_TAMOS, PolicyConstants.ACL_TYPE_NONE)  : NoACL.convertFromTAMOS,
             (PolicyConstants.ACL_TYPE_TAMOS, PolicyConstants.ACL_TYPE_NFSV4) : NFSv4.convertFromTAMOS,
             (PolicyConstants.ACL_TYPE_TAMOS, PolicyConstants.ACL_TYPE_AIXC)  : AIXC.convertFromTAMOS,
             
             (PolicyConstants.ACL_TYPE_NONE, PolicyConstants.ACL_TYPE_POSIX)  : POSIX.convertFromNone,
             (PolicyConstants.ACL_TYPE_NONE, PolicyConstants.ACL_TYPE_NFSV4)  : NFSv4.convertFromNone,
             (PolicyConstants.ACL_TYPE_NONE, PolicyConstants.ACL_TYPE_AIXC)   : AIXC.convertFromNone,

             (PolicyConstants.ACL_TYPE_AIXC, PolicyConstants.ACL_TYPE_NFSV4)  : NFSv4.convertFromAIXC
        }

        # valid special-bits settings
        self.__validSpecialBits = [
             None,
             '',
             'none',
             'setuid',
             'setgid',
             'sticky',
             'setgid,sticky',
             'setuid,sticky',
             'setuid,setgid',
             'setuid,setgid,sticky'
        ]
        
    def initialize(self, target):
        self.target = target
        
        # set up a logger, mostly for trace purposes
        self.logger = Logger.getLogger(__name__)
        if target.getLogger() is not None:
            self.logger.setParent(target.getLogger())
        
        
    def shutdown(self):
        # hm
        pass

    def toString(self):
        return "Filesystem ACL"

    def compareAttribute(self, result, obj, attr, value):
        """
        Returns True if the value of the attribute on the object does not 
        match the expected value, as long as one is provided.  If 'value' is
        None, always returns False.
        """
        objValue = obj.getExtAttributeValue(attr, 0)
        if value is None or value == objValue:
            return False
        else:
            result.diff(attr, None, objValue, value)
            return True

    def parseToolError(self, result, err):
        """
        To avoid making the ACL tool too complicated, it doesn't format error messages itself.
        Instead, it spits out a limited form of error message for each thing that goes wrong,
        which we convert to human readable messages here.
        """
        errors = str(err)
        exc = None
        warnusers = {}
        warngroups = {}
        for e in errors.split("\n"):
            if e == "":
                continue

            elif e.startswith("u:"):
                # so far, always means 'user doesn't exist'
                username = e[2:]
                if username not in warnusers:
                    result.actionWarning(npm_m_user_does_not_exist, (username,))
                    warnusers[username] = 1

            elif e.startswith("g:"):
                # so far, always means 'group doesn't exist'
                groupname = e[2:]
                if groupname not in warngroups:
                    result.actionWarning(npm_m_group_does_not_exist, (groupname,))
                    warngroups[groupname] = 1

            elif e.startswith("c:"):
                try:
                    (action, errname, errstring) = e[2:].split(":", 3)

                    # some errors, we don't really care what caused it, but the specific error is interesting
                    if errname == 'EROFS':
                        exc = PluginException(self, npm_m_target_filesystem_read_only)
                    elif errname == 'EACCES' or errname == 'EPERM':
                        # this should never happen
                        exc = PluginException(self, npm_m_target_permission_denied)
                    elif errname == 'ENOTSUP' or errname == 'ENOSYS':
                        exc = PluginException(self, npm_m_target_filesystem_no_acl_support)
                    elif errname == 'ENOENT':
                        exc = PluginException(self, npm_m_target_file_does_not_exist)
                    elif action == 'open' or action == 'stat':
                        exc = PluginException(self, npm_m_target_unable_to_open_file, (errstring,))
                    elif action == 'chown':
                        exc = PluginException(self, npm_m_target_unable_to_update_ownership, (errstring,))
                    elif action == 'chmod':
                        exc = PluginException(self, npm_m_target_unable_to_update_special_bits, (errstring,))
                    elif action == 'getacl':
                        exc = PluginException(self, npm_m_target_unable_to_retrieve_file_acl, (errstring,))
                    elif action == 'setacl':
                        exc = PluginException(self, npm_m_target_unable_to_apply_file_acl, (errstring,))
                    else:
                        exc = PluginException(self, npm_m_target_generic_system_error, (errstring,))
                except:
                    exc = PluginException(self, npm_m_target_generic_system_error, (str(e),))

            elif e.startswith("t:"):
                # this indicates an unsupported ACL entry type was passed down
                result.actionWarning(npm_m_unsupported_acl_entry_type, (e[2:],))

            elif e.startswith("e:"):
                # this indicates a generic error message
                exc = PluginException(self, npm_m_target_generic_system_error, (e[2:],))

        if exc is not None:
            raise exc

    def getState(self, result, actualPath):
        """
        Reads the current state of a target file, returning File and ACL objects.
        """
        result.newStep("Reading state of file " + actualPath)
        
        out = StringBuffer()
        err = StringBuffer()
        rc = self.target.execute(("/opt/pdos/sbin/pdosfsacl", actualPath, "get"), None, out, err)
        if err.length() != 0:
            self.parseToolError(result, err)
        
        file = File()
        acl = ACL()
        aclin = LineNumberReader(StringReader(out.toString()))
        try:
            fileType = aclin.readLine()
            ownerInfo = aclin.readLine()
            specialBits = aclin.readLine()
            
            file.addExtAttribute(PolicyConstants.UNIX_FILE_TYPE, fileType)
            (owner, owningGroup) = ownerInfo.split(":")
            file.addExtAttribute(RESOURCE_OWNER, owner)
            file.addExtAttribute(RESOURCE_OWNING_GROUP, owningGroup)
        
            file.addExtAttribute(UNIX_FILE_SPECIAL_BITS, specialBits)

            supportedAclTypes = aclin.readLine()
            for supportedAclType in supportedAclTypes.split(","):
                file.addExtAttribute("supported_acl_type", supportedAclType)
            
            aclType = aclin.readLine()
            if aclType not in self.__formats:
                raise IllegalStateException("Unsupported ACL format " + aclType)
            self.__formats[aclType].convertToACL(result, acl, aclin)
        except java.io.IOException, e:
            raise PluginException(self, npm_m_target_generic_system_error, str(e))
        
        acl.setKey("ACL from " + actualPath)
        file.setKey(actualPath)
        return (file, acl)


    def determineOwnership(self, result, acl, file, warn=True):
        """
        Determines the effective policy regarding ownership of a file, given
        the policy file object and the ACL that applies to it.  Optionally issues
        a warning if the owner or owning group is specified in both.
        """
        fileOwner = file.getExtAttributeValue(RESOURCE_OWNER, 0)
        fileOwningGroup = file.getExtAttributeValue(RESOURCE_OWNING_GROUP, 0)
        aclOwner = acl.getExtAttributeValue(RESOURCE_OWNER, 0)
        aclOwningGroup = acl.getExtAttributeValue(RESOURCE_OWNING_GROUP, 0)

        if warn:
            if fileOwner and aclOwner:
                result.actionWarning(npm_m_file_ownership_on_file_and_acl, (file.getKey(), acl.getKey()))
            if fileOwningGroup and aclOwningGroup:
                result.actionWarning(npm_m_file_group_ownership_on_file_and_acl, (file.getKey(), acl.getKey()))

        o = (fileOwner or aclOwner, fileOwningGroup or aclOwningGroup)
        self.logger.finer("Ownership of file %s: %s" % (file.getKey(), o))
        return o

    

    def applyState(self, result, actualPath, file, acl):
        """
        Applies information from 'file' and 'acl' to the filesystem,
        using the pdosfsacl tool.
        """
        result.newStep("Applying ACL to file " + actualPath)
        
        # check ACL format
        aclFormat = acl.getSource()

        if aclFormat not in self.__setableFormats:
            raise IllegalStateException("Unsupported ACL format: " + aclFormat)

        (owner, group) = self.determineOwnership(result, acl, file, False)

        args = ["/opt/pdos/sbin/pdosfsacl", actualPath, "set"]
        args.append(owner or "")
        args.append(group or "")
        args.append(file.getExtAttributeValue(UNIX_FILE_SPECIAL_BITS, 0) or "")

        args += self.__formats[aclFormat].convertToArgs(result, acl)

        out = StringBuffer()
        err = StringBuffer()
        rc = self.target.execute(args, None, out, err)
        if err.length() != 0:
            self.parseToolError(result, err)


    def determineACLFormat(self, result, targetFile):
        """
        Determines which ACL type to set for the target file
        """
        supportedAclTypes = targetFile.getExtAttribute("supported_acl_type")
        for priority in self.__setableFormats:
            if priority in supportedAclTypes:
                return priority

        return None

        
    def convertPolicyACL(self, result, targetFile, targetAcl, policyFile, policyAcl):
        """
        Converts an ACL object from the policy into the format supported by the target filesystem
        """
        resultAclFormat = self.determineACLFormat(result, targetFile)
        if resultAclFormat is None or resultAclFormat not in self.__formats:
            raise IllegalStateException("Do not know how to set any of the ACL types for file %s (%s)" % (policyFile.getKey(), targetFile.getExtAttribute("supported_acl_type")))

        if resultAclFormat == targetAcl.getSource():
            baseAcl = targetAcl
        else:
            baseConverterKey = (targetAcl.getSource(), resultAclFormat)
            if baseConverterKey not in self.__converters:
                raise IllegalStateException("Don't know how to convert ACL format %s from policy for filesystem" % (PolicyConstants.ACL_TYPE_NONE, resultAclFormat))
            baseAcl = self.__converters[baseConverterKey](result, targetFile, targetAcl, targetFile, targetAcl)
            
        converterKey = (policyAcl.getSource(), resultAclFormat)
        if converterKey not in self.__converters:
            raise IllegalStateException("Don't know how to convert ACL format %s from policy to %s for filesystem" % (policyAclFormat, resultAclFormat))
        
        # perform the conversion
        resultAcl = self.__converters[converterKey](result, targetFile, baseAcl, policyFile, policyAcl)
        return resultAcl



    def extractWildcard(self, result, wildcards, file):
        """
        If the file is a wildcard specification, adds it to the wildcard map
        and returns None, otherwise returns the filename with any escaped
        wildcard characters unescaped.
        """

        (dir, name, regex) = ossutil.wildcard.extract(file.getKey())
        if dir is None:
            if result is not None:
                result.startAction(file)
                result.finishAction(npm_m_file_spec_includes_non_leaf_wildcards, [file.getKey()])
            return None

        if regex is not None:
            if dir not in wildcards:
                wildcards[dir] = []
            wildcards[dir].append([file, name, regex])

            return None
        else:
            return dir + "/" + name



    def expandWildcard(self, result, dir, policy):
        """
        Matches directory contents against wildcard specifications, returning a mapping
        from filename => file object.
        dir: directory name (untranslated)
        policy: the policy from which to extract further wildcard matches for the directory
        """

        # get other (less interesting) wildcard patterns for this directory
        rq = ResourceQuery(PolicyConstants.FILE)
        rq.resourcePrefixQuery(PolicyDelta.ALL, dir)

        # hack
        pd = PolicyDelta()
        pd.newPolicy = policy
        pd.oldPolicy = policy
        files = rq.evaluate(pd)

        wildcards = {}
        for f in files:
            (fdir, fname) = ossutil.splitDirName(f.getKey())
            if fdir != dir:
                # wrong directory
                continue

            if self.extractWildcard(None, wildcards, f) is not None:
                # this was a file
                continue


        if dir not in wildcards:
            # no wildcards
            return []

        # okay, now we have a full set of the wildcard patterns that apply to this directory.
        # sort them by priority
        patterns = wildcards[dir]
        def compare_regexp(a, b):
            (awcf, apat, areg) = a
            (bwcf, bpat, breg) = b
            return ossutil.wildcard.compare_regexp(apat, bpat)

        patterns.sort(compare_regexp)

        # compile the regular expressions
        cpatterns = []
        for (wcf, pat, reg) in patterns:
            cpatterns.append([wcf, pat, re.compile(reg)])

        # get the list of files in the directory
        # (we don't even know if it is a directory until we get here)
        actualDir = self.target.translatePath(dir)
        out = java.lang.StringBuffer()
        err = java.lang.StringBuffer()
        try:
            files = self.target.execute(["/opt/pdos/sbin/pdosfsacl", actualDir, "list"], None, out, err)
            if err.length() != 0:
                self.parseToolError(result, err)
        except java.lang.Exception, e:
            result.startAction(File(dir, None, None))
            result.finishAction(e)
            return []
        except Exception, e:
            result.startAction(File(dir, None, None))
            result.finishAction(wrapPyException(self, e))
            return []

        
        results = {}
        flist = LineEditor(out.toString())
        for f in flist.rawlines():
            # check if there's a policy object for this exact file
            ffn = ossutil.wildcard.escape(dir + "/" + f)
            if ffn in policy.resourceMap[PolicyConstants.FILE]:
                # probably record this somewhere
                continue

            # find the most specific pattern that applies to the file
            for (wcf, pat, reg) in cpatterns:
                if reg.match(f):
                    fullname = dir + "/" + f
                    results[fullname] = wcf
                    break

        return results





    def updateACL(self, result, policy, file, filename, apply, deleted=False):
        """
        Processes a single file, ensuring that its ACL is up to date with the current policy
        """
        result.startAction(file)
        outcome = npm_s_success
        try:
            if deleted or file.getACLID() is None:
                acl = ACL()
                acl.setKey("No ACL")
                acl.setSource(PolicyConstants.ACL_TYPE_NONE)
                for a in (RESOURCE_OWNER, RESOURCE_OWNING_GROUP, UNIX_FILE_SPECIAL_BITS):
                    if file.getExtAttributeValue(a, 0) is not None:
                        break
                else:
                    if deleted and file.getACLID() is None:
                        # file doesn't have any policy on it, and we don't need to process it
                        # regardless, so ignore it.
                        result.finishAction(outcome, [])
                        return
            else:
                acl = policy.getRule(PolicyConstants.ACL, file.getACLID())
                if acl is None:
                    raise IllegalStateException("File %s has unknown ACL %s attached" % (file.getKey(), file.getACLID()))

                result.addRule(acl)

            actualPath = self.target.translatePath(filename)
            
            # get current state
            (targetFile, targetAcl) = self.getState(result, actualPath)

            applyAcl = self.convertPolicyACL(result, targetFile, targetAcl, file, acl)
            aclFormat = applyAcl.getSource()

            (policyOwner, policyOwningGroup) = self.determineOwnership(result, applyAcl, file)
            policySpecialBits = file.getExtAttributeValue(UNIX_FILE_SPECIAL_BITS, 0)
            if policySpecialBits not in self.__validSpecialBits:
                result.actionWarning(npm_m_unknown_file_special_bits, (policySpecialBits,))
                policySpecialBits = None

            different = False
            different = self.compareAttribute(result, targetFile, RESOURCE_OWNER, policyOwner) or different
            different = self.compareAttribute(result, targetFile, RESOURCE_OWNING_GROUP, policyOwningGroup) or different
            different = self.compareAttribute(result, targetFile, UNIX_FILE_SPECIAL_BITS, policySpecialBits) or different

            if targetAcl.getSource() != applyAcl.getSource() or \
                (self.__formats[aclFormat].compareACLs(result, targetAcl, applyAcl) == False):
                different = True
                # in addition to the individual differences, show the whole ACL as changing:
                result.diff("acl", None, targetAcl, applyAcl)

            if different:
                if apply:
                    self.applyState(result, actualPath, file, applyAcl)
                else:
                    outcome = npm_s_policy_not_compliant

            result.finishAction(outcome, [])
        except java.lang.Exception, e:
            result.finishAction(e)
        except Exception, e:
            result.finishAction(wrapPyException(self, e))


    def policyUpdate(self, result, delta, apply):
        """
        Processes an update to the policy.
        """
        
        # find newly added and updated files, as well as files for which
        # the attached ACL has been updated
        rq = ResourceQuery(PolicyConstants.FILE)
        rq.resourceQuery(PolicyDelta.ADDED)
        rq.resourceQuery(PolicyDelta.CHANGED)
        rq.resourceRuleQuery(PolicyDelta.CHANGED, PolicyConstants.ACL)
        files = rq.evaluate(delta)
        wildcardDirs = {}
        for f in files:
            filename = self.extractWildcard(result, wildcardDirs, f)
            if filename is not None:
                # not a wildcard, so apply it directly
                self.updateACL(result, delta.newPolicy, f, filename, apply)

        for dir in wildcardDirs:
            # get the set of patterns we care about
            interesting = [file for (file, pat, reg) in wildcardDirs[dir]]
            expanded = self.expandWildcard(result, dir, delta.newPolicy)
            for filename in expanded:
                # if the file policy object for this filename is in the modified set, apply the ACL
                if expanded[filename] in interesting:
                    self.updateACL(result, delta.newPolicy, expanded[filename], filename, apply)

        # find files that have been removed and remove the ACL entries from the filesystem.
        # for each directory in which files or wildcards were removed, reapply all wildcards
        # to ensure the policy is correctly applied.
        rq = ResourceQuery(PolicyConstants.FILE)
        rq.resourceQuery(PolicyDelta.DELETED)
        files = rq.evaluate(delta)
        wildcardDirs = []
        for f in files:
            (dir, name, regex) = ossutil.wildcard.extract(f.getKey())
            if dir is None:
                # bad wildcard spec was removed.  don't annoy anyone about it 
                # any further.
                continue

            wildcardDirs.append(dir)
            if name is not None:
                fullname = dir + "/" + name
                self.updateACL(result, delta.oldPolicy, f, fullname, apply, deleted=True)

        for dir in wildcardDirs:
            expanded = self.expandWildcard(result, dir, delta.newPolicy)
            for filename in expanded:
                self.updateACL(result, delta.newPolicy, expanded[filename], filename, apply)

            # now find files that are no longer covered by policy
            oldExpanded = self.expandWildcard(result, dir, delta.oldPolicy)
            for filename in oldExpanded:
                if filename not in expanded:
                    self.updateACL(result, delta.newPolicy, oldExpanded[filename], filename, apply, deleted=True)

        result.finish(npm_s_success, [])


    def verifyACL(self, result, policy, file, filename, reconcile):
        """
        Checks that the current filesystem ACL on the file matches the policy.
        If 'reconcile' is True, updates the filesystem ACL to comply with the policy.
        """
        result.startAction(file)
        outcome = npm_s_success
        try:
            if file.getACLID() is None:
                acl = ACL()
                acl.setSource(PolicyConstants.ACL_TYPE_NONE)
                acl.setKey("No ACL")
            else:
                acl = policy.getRule(PolicyConstants.ACL, file.getACLID())
                if acl is None:
                    raise IllegalStateException("File %s has unknown ACL %s attached" % (file.getKey(), file.getACLID()))

            result.addRule(acl)

            actualPath = self.target.translatePath(filename)

            # get current state
            (targetFile, targetAcl) = self.getState(result, actualPath)
            convertedAcl = self.convertPolicyACL(result, targetFile, targetAcl, file, acl)
            targetFormat = targetAcl.getSource()

            (policyOwner, policyOwningGroup) = self.determineOwnership(result, convertedAcl, file)
            policySpecialBits = file.getExtAttributeValue(UNIX_FILE_SPECIAL_BITS, 0)
            if policySpecialBits not in self.__validSpecialBits:
                result.actionWarning(npm_m_unknown_file_special_bits, (policySpecialBits,))
                policySpecialBits = None

            apply = False
            apply = self.compareAttribute(result, targetFile, RESOURCE_OWNER, policyOwner) or apply
            apply = self.compareAttribute(result, targetFile, RESOURCE_OWNING_GROUP, policyOwningGroup) or apply
            apply = self.compareAttribute(result, targetFile, UNIX_FILE_SPECIAL_BITS, policySpecialBits) or apply

            if targetFormat != convertedAcl.getSource() or \
                (self.__formats[targetFormat].compareACLs(result, targetAcl, convertedAcl) == False):
                # results.diff("acl", None, targetAcl, convertedAcl)
                apply = True

            if apply:
                if reconcile:
                    self.applyState(result, actualPath, file, convertedAcl)
                    outcome = npm_s_policy_reconciled
                else:
                    result.trace("File %s does not comply with current policy" % (filename))
                    outcome = npm_s_policy_not_compliant
            else:
                result.trace("File %s complies with current policy" % (filename))
            
            result.finishAction(outcome, [])
        except java.lang.Exception, e:
            result.finishAction(e)
        except Exception, e:
            result.finishAction(wrapPyException(self, e))

        return outcome


        

    def verifyPolicy(self, result, policy, reconcile):
        """
        Checks that the ACLs on all files described by the policy are 
        in place, reporting on differences and (optionally) reconciling them.
        """
    
        self.logger.fine("Performing policy verification")
        files = policy.resourceMap[PolicyConstants.FILE]
        if files is None:
            result.finish(npm_s_success, [])
            return
    
        outcome = npm_s_success
        wildcardDirs = {}
        for file in files.values():
            filename = self.extractWildcard(result, wildcardDirs, file)
            if filename is not None:
                self.logger.finer("Verifying filesystem ACL policy for file %s" % (filename))
                o = self.verifyACL(result, policy, file, filename, reconcile)
                if outcome == npm_s_success:
                    outcome = o

        for dir in wildcardDirs:
            expanded = self.expandWildcard(result, dir, policy)
            for file in expanded:
                o = self.verifyACL(result, policy, expanded[file], file, reconcile)
                if outcome == npm_s_success:
                    outcome = o


        result.finish(outcome, [])
        

# vim: set ts=4 sw=4 expandtab :

