#
# IBM Confidential
# OCO Source Materials
# Tivoli Access Manager for Operating Systems - Agent
# TIVO-TAMOSA
#
# (C) Copyright IBM Corp. 2008
# The source code for this program is not published or otherwise
# divested of its trade secrets, irrespective of what has been
# deposited with the U.S. Copyright Office.
#

#
# aix login policy implementation
#

from com.ibm.amos.agent.db import PolicyConstants
from com.ibm.amos.agent.db.resources import User
from com.ibm.amos.agent.db.resources import File
from com.ibm.amos.agent.db.rules import LoginActivity
from com.ibm.amos.nls.pdopmidx import *

import java.lang.Exception
import java.lang.String
import java.lang.StringBuffer
import java.util.logging.Level

from ossutil import wrapPyException
from ossutil.configeditor import StanzaFile
from ossutil.aixconfig import AixSystemConfigFile

LOGIN_ACTIVITY_CONFIG_FILE = "/etc/security/user"

MIN_AGE = "minage"
MAX_AGE = "maxage"
LOGIN_RETRIES = "loginretries"
DEFAULT_USER = "default"

CONTROLLED_VALUES = ( MIN_AGE, MAX_AGE, LOGIN_RETRIES )

class AixLoginActivityPlugin:
    def __init__(self, realplugin, target):
        self.target = target
        self.realplugin = realplugin
        self.__l = target.logger

    def shutdown(self):
        # do nothing
        pass

    def applyPolicy(self, result, policy, update, warnOnly = False):
        """
        Apply a policy, either as an update or a reconcilliation, possibly
        not even applying the policy.
        """
        outcome = npm_s_success
        try:
            # Read the current state
            xau = self.target.translatePath(LOGIN_ACTIVITY_CONFIG_FILE)
            data = java.lang.String(self.target.getFile(xau))
            loginFile = AixSystemConfigFile(data)
            
            outcome = self.applyPolicyImpl(result, loginFile, policy, update, warnOnly)

            if loginFile.hasChanged():
                result.startAction(File(LOGIN_ACTIVITY_CONFIG_FILE, None, None))
                result.newStep("Updating the login activity config file")
                self.__l.finest("New config file contents:\n%s\n" % (loginFile.getText()))
                self.target.putFile(xau, loginFile.getText(), True)
		result.finishAction(npm_s_success, [])

            result.finish(outcome, [])

        except java.lang.Exception, e:
            result.finish(e)
        except Exception, e:
            result.finish(wrapPyException(self.realplugin, e))


    def determineLoginActivityUsers(self, target, policy):
        """
        Figures out which login activity settings must be applied to which users.
        """
        users = {}

        def processUser(users, name, lap):

            # TAMOS specifies min / max age values in days, AIX specifies them
            # in weeks. For both min / max age, we round the value down to a
            # minimum of 1 (for non-zero values)
            def age_conversion(value_str):
                value = int(value_str)
                if value < 1:
                    value = 0
                else:
                    value = max(1, value / 7)
                return value

            minage_str = lap.getMinPasswordDays()
            maxage_str = lap.getMaxPasswordDays()
            retries_str = lap.getMaxFailedLogins()

            disabled_str = None
            if not name == DEFAULT_USER:
                disabled_str = lap.getPolicyDisabled()

            if minage_str is not None or maxage_str is not None or retries_str is not None or disabled_str is not None:
                if not users.has_key(name):
                    users[name] = {}

            if disabled_str is not None:
                users[name][MIN_AGE] = 0
                users[name][MAX_AGE] = 0
                users[name][LOGIN_RETRIES] = 0
            else:
                if minage_str is not None:
                    users[name][MIN_AGE] = age_conversion(minage_str)
                if maxage_str is not None:
                    users[name][MAX_AGE] = age_conversion(maxage_str)
                if retries_str is not None:
                    users[name][LOGIN_RETRIES] = max(0, int(retries_str))


        userRes = policy.resourceMap[PolicyConstants.USER]
        if userRes is not None:
            for name in userRes:
                if name == DEFAULT_USER:
                    # Log a warning about this...
                    
                    pass
                else:
                    user = userRes[name]
                    user_lap_id = user.getLoginID()
                    if user_lap_id is not None:
                        user_lap = policy.getRule(PolicyConstants.LOGIN_ACTIVITY, user_lap_id)
                        if user_lap is not None:
                            processUser(users, name, user_lap)

        # Find if we have a global policy to apply
        global_lap = None
        global_res = policy.resourceMap[PolicyConstants.GLOBAL]
        if global_res is not None:
            if len(global_res) == 1:
                for name in global_res:
                    elt = global_res[name]
                    global_lap_id = elt.getLoginID()
                    if global_lap_id is not None:
                        global_lap = policy.getRule(PolicyConstants.LOGIN_ACTIVITY, global_lap_id)
			# ignore the global policy if no attributes are set
			if global_lap == LoginActivity(None, None):
			    global_lap = None
            else:
                self.__l.finer("Unexpected number of global elements found.")
                

        # If we have global login activity policy, or we have user specific
        # overrides, then we should ensure that we have default values for
        # each of the entries that we control
        if global_lap is not None or len(users) > 0:
            users[DEFAULT_USER] = {}
            users[DEFAULT_USER][MIN_AGE] = 0
            users[DEFAULT_USER][MAX_AGE] = 0
            users[DEFAULT_USER][LOGIN_RETRIES] = 0

        if global_lap is not None:
            processUser(users, DEFAULT_USER, global_lap)

        return users

    def applyPolicyImpl(self, result, loginFile, policy, update, warnOnly = False):
        """
        Implementation of actual policy application. This has been implemented
        as a separate function mainly for ease of unit testing
        """
        outcome = npm_s_success
        
        loginFile.parse()
        users = self.determineLoginActivityUsers(self.target, policy)

        def updateCb(stanzaName, name, values, data):
            (result, users) = data
            result.startAction(User(name, "TAMOS"))

            if name in CONTROLLED_VALUES:
                retval = None
            else:
                retval = values

            if users.has_key(stanzaName):
                if users[stanzaName].has_key(name):
                    retval = ["%d" % (users[stanzaName][name])]
                    del users[stanzaName][name]
                    if len(users[stanzaName].keys()) == 0:
                        del users[stanzaName]

            if not retval == values:
                if retval is None:
                    result.diff("user", name, ",".join(values), None)
                else:
                    result.diff("user", name, ",".join(values), ",".join(retval))

            if update or retval == values:
                result.finishAction(npm_s_success, [])
            elif warnOnly:
                result.finishAction(npm_s_policy_not_compliant, [])
                retval = values
            else:
                result.finishAction(npm_s_policy_reconciled, [])

            return retval

        # If there is no login activity policy set, we assume this means that
        # they are not using TAMOS to centrally manage login activity policy,
        # so we leave any existing policy unmodified.
        if len(users.keys()) > 0:
            self.__l.finer("Updating any existing login activity policy settings")
            loginFile.updateAllStanzas(updateCb, (result, users))

            self.__l.finer("Adding any remaining users with no login activity policy settings")
            for user in users:
                for elt in users[user]:

                    result.startAction(User(user, "TAMOS"))
                    values = ["%d" % (users[user][elt])]

                    result.diff("user", user, None, ",".join(values))
                    if update:
                        loginFile.appendEntry(user, elt, values)
                        result.finishAction(npm_s_success, [])
                    else:
                        if warnOnly:
                            result.finishAction(npm_s_policy_not_compliant, [])
                        else:
                            loginFile.appendEntry(user, elt, values)
                            result.finishAction(npm_s_policy_reconciled, [])
        else:
            self.__l.finer("No login activity policy has been defined, leaving existing configuration unchanged.")

        return outcome
        
    
    def policyUpdate(self, result, delta, apply):
        """
        However, because w have to manually expand the groups, we apply
        the full policy every time, since the delta may not actually pick
        up all the changes. Well, that was the theory, but we're no longer
        supporting groups for login activity policy.
        """
        self.applyPolicy(result, delta.newPolicy, True, False)

    def verifyPolicy(self, result, policy, reconcile):
        """
        Verify that the policy that is currently on the system complies with
        what is set in TAMOS. Optionally, update the policy to reflect with
        what TAMOS has. Well, that was the theory, but we're no longer
        supporting groups for login activity policy.
        """
        self.applyPolicy(result, policy, False, not reconcile)

