#
# IBM Confidential
# OCO Source Materials
# Tivoli Access Manager for Operating Systems - Agent
# TIVO-TAMOSA
#
# (C) Copyright IBM Corp. 2008
# The source code for this program is not published or otherwise
# divested of its trade secrets, irrespective of what has been
# deposited with the U.S. Copyright Office.
#

#
# solaris login activity implementation
#

from __future__ import generators

from com.ibm.amos.agent.db import PolicyConstants
from com.ibm.amos.agent.db.rules import LoginActivity
from com.ibm.amos.agent.db.resources import Global, User
from com.ibm.amos.nls.pdopmidx import *

from ossutil import escapeSplit, unescape, escapeJoin
from ossutil.keyvalueeditor import KeyValueEditor
from ossutil.configeditor import LineEditor
from ossutil.solaris import SolarisShadowFile

import java.lang.String

LOGIN_DEFAULTS = "/etc/default/login"
PASSWD_DEFAULTS = "/etc/default/passwd"
SECURITY_POLICY_CONF = "/etc/security/policy.conf"
USER_ATTR = "/etc/user_attr"
SHADOW = "/etc/shadow"


class SolarisUserAttrFile:
    """
    Specialised editor for the Solaris /etc/user_attr file
    """
    def __init__(self, data):
        self.edit = LineEditor(java.lang.String(data), has_comments=True, has_continuations=True)

    def __split(self, line):
        try:
            (username, qualifier, res1, res2, attr) = line.split(":", 5)
            escattrs = escapeSplit(attr, ";")
            # qualifier, res1, and res2 are all uninteresting
            return (username, map(unescape, escattrs))
        except ValueError:
            # broken line
            return None

    def entries(self):
        for line in self.edit.lines():
            r = self.__split(line)
            if r is not None:
                yield r

    def __format(self, username, attrs):
        return "%s::::%s" % (username, escapeJoin(attrs, ";", ":\\"))

    def update(self, callback, cbdata):
        def cb(line, (callback, data)):
            r = self.__split(line)
            if r is None:
                return line

            updated = callback(r[0], r[1], data)
            if updated is None:
                return None

            (username, attrs) = updated
            if username == "" or len(attrs) == 0:
                return None
            
            return self.__format(username, attrs)

        self.edit.update(cb, (callback, cbdata))

    def append(self, username, attrs):
        self.edit.append(self.__format(username, attrs))

    def hasChanged(self):
        return self.edit.hasChanged()

    def getData(self):
        return self.edit.getData()



class SolarisLoginActivityPlugin:
    def __init__(self, realplugin, target):
        self.target = target
        self.realplugin = realplugin
        
    def shutdown(self):
        pass

    def _setKeys(self, result, path, keys, apply=True, has_comments=True):
        xpath = self.target.translatePath(path)
        fd = self.target.getFile(xpath)

        ke = KeyValueEditor(java.lang.String(fd), has_comments=has_comments)
        for k in keys:
            current = ke.get(k)
            if current != keys[k]:
                result.diff(path, k, current, keys[k])
                ke.set(k, keys[k])

        if ke.hasChanged() and apply:
            self.target.putFile(xpath, ke.getData(), True)

        return ke.hasChanged()

    def setRetries(self, result, retries, apply):
        # maximum value for RETRIES is 15
        if retries is not None and int(retries) > 15:
            retries = 15
        return self._setKeys(result, LOGIN_DEFAULTS, { "RETRIES": retries }, apply)

    def setLockAfterRetries(self, result, lock, apply):
        return self._setKeys(result, SECURITY_POLICY_CONF, { "LOCK_AFTER_RETRIES": lock }, apply)

    def clearPasswordAge(self, result, apply):
        return self._setKeys(result, PASSWD_DEFAULTS, { "MINWEEKS": None, "MAXWEEKS": None }, apply)


    def getGlobalPolicy(self, policy):
        if policy is None:
            return None

        g = policy.getResource(PolicyConstants.GLOBAL, "GLOBAL")
        if g is not None and g.getLoginID() is not None:
            gp = policy.getRule(PolicyConstants.LOGIN_ACTIVITY, g.getLoginID())
            # ignore it if no attributes are set
            if gp != LoginActivity("nothing", None):
                return gp

        return None

    def getUserPolicy(self, policy):
        # find users with attached login policy
        userpolicy = {}
        users = policy.resourceMap.get(PolicyConstants.USER) or {}
        loginpolicy = policy.ruleMap.get(PolicyConstants.LOGIN_ACTIVITY) or {}
        for username in users:
            user = users.get(username)
            if user is None or user.getLoginID() is None:
                continue

            up = loginpolicy.get(user.getLoginID())
            userpolicy[username] = up

        return userpolicy


    def applyGlobalMaxFailures(self, user):
        """
        Global maximum login failure policy doesn't apply to some users,
        notably root, in order to avoid locking the entire system.
        """
        return user != 'root'

    def updateShadowRecord(self, record, (result, globalpolicy, userpolicy, currentUserLock, updateUserLock, minUserFail)):
        """
        Processes a record in /etc/shadow, updating the min, max, and inactive fields,
        and figuring out if the user needs a lock_after_retries attribute set in /etc/user_attr.
        """
        user = record['username']
        result.startAction(User(user, "TAMOS"))
        result.addRule(globalpolicy)

        up = None
        globaldisabled = globalpolicy.getPolicyDisabled() not in (None, "")
        disabled = globaldisabled

        if user in userpolicy:
            up = userpolicy[user]
            result.addRule(up)
            disabled = globaldisabled or up.getPolicyDisabled() not in (None, "")

        # apply settings we manage in /etc/shadow
        for (field, getfield) in (('min', LoginActivity.getMinPasswordDays),
                                  ('max', LoginActivity.getMaxPasswordDays),
                                  ('inactive', LoginActivity.getMaxInactiveDays)):
            value = None
            if not disabled:
                # check user policy
                if up is not None:
                    value = getfield(up)

                # fall back to global policy
                if value is None:
                    value = getfield(globalpolicy)

            cv = record[field] or None  # "" vs None
            if value != cv:
                result.diff(SHADOW, field, cv, value)
                record[field] = value or ""


        # figure out if we need to update /etc/user_attr for this user, so we can report it
        # as part of the same action
        expectLock = None
        globalfail = globalpolicy.getMaxFailedLogins()

        if disabled:
            if globaldisabled is False and globalfail is not None and int(globalfail) > 0:
                # global maxFailedLogins is nonzero, so we need to set lock_after_retries=no
                # for this user
                expectLock = "no"
        elif up is not None:

            userfail = up.getMaxFailedLogins()

            # figure out the lowest non-zero number of failures allowed
            if userfail is not None:
                uf = int(userfail)
                if uf > 0 and (minUserFail[0] == 0 or uf < minUserFail[0]):
                    minUserFail[0] = uf

            if self.applyGlobalMaxFailures(user) is False and userfail is None:
                expectLock = "no"
            elif (globalfail is not None and int(globalfail) > 0) and (userfail is not None and int(userfail) == 0):
                # global maxFailedLogins is nonzero, user's maxFailedLogins is zero, so
                # they need to have lock_after_retries=no
                expectLock = "no"
            elif (globalfail is None or int(globalfail) == 0) and (userfail is not None and int(userfail) > 0):
                # global maxFailedLogins is unset or zero, user's maxFailedLogins is nonzero,
                # so they need to have lock_after_retries=yes
                expectLock = "yes"
        elif self.applyGlobalMaxFailures(user) is False:
            # only enable account locking for root if the policy specifically says to
            expectLock = "no"

        if currentUserLock.get(user) != expectLock:
            result.diff(USER_ATTR, "lock_after_retries", currentUserLock.get(user), expectLock)
            updateUserLock[user] = expectLock

        result.finishAction(npm_s_success, [])
        return record



    def applyLoginPolicy(self, result, globalpolicy, userpolicy, apply):
        # update /etc/shadow and /etc/user_attr
        shadowpath = self.target.translatePath(SHADOW)
        shadow = SolarisShadowFile(self.target.getFile(shadowpath))
        
        userattrpath = self.target.translatePath(USER_ATTR)
        userattr = SolarisUserAttrFile(self.target.getFile(userattrpath))

        currentUserLock = {}        # current state of lock_after_retries in /etc/user_attr
        updateUserLock = {}         # changes we're going to make to lock_after_retries in /etc/user_attr
        minUserFail = [0]           # lowest non-zero user value for maxFailedLogins
        changed = False

        # figure out the current user-specific lock-after-retries settings, which we
        # update while processing the shadow file records
        for (username, attrs) in userattr.entries():
            if "lock_after_retries=yes" in attrs:
                currentUserLock[username] = "yes"
            elif "lock_after_retries=no" in attrs:
                currentUserLock[username] = "no"

        # update /etc/shadow (setting min, max, inactive) and figure out what
        # we need to do to /etc/user_attr.
        shadow.update(self.updateShadowRecord, (result, globalpolicy, userpolicy, currentUserLock, updateUserLock, minUserFail))

        def userattrcb(username, attrs, updateUserLock):
            """
            Updates /etc/user_attr records, setting lock_after_retries where required.
            """
            if username in updateUserLock:
                attrs = [ attr for attr in attrs if attr.startswith("lock_after_retries=") == False ]
                if updateUserLock[username] is not None:
                    attrs.append("lock_after_retries=%s" % updateUserLock[username])
                del updateUserLock[username]

            return (username, attrs)

        userattr.update(userattrcb, updateUserLock)

        # now add any extra users we might need
        for u in updateUserLock:
            userattr.append(u, ["lock_after_retries=%s" % updateUserLock[u]])

        # update /etc/shadow and /etc/user_attr files if we've changed them
        if shadow.hasChanged() and apply:
            self.target.putFile(shadowpath, shadow.getData(), True)
            changed = True

        if userattr.hasChanged() and apply:
            self.target.putFile(userattrpath, userattr.getData(), True)
            changed = True

        # now apply global login policy
        result.startAction(Global("global", "TAMOS"))
        result.addRule(globalpolicy)

        # always clear the password age settings in /etc/default/passwd
        changed = self.clearPasswordAge(result, apply) or changed

        # set the number of login failures allowed
        gf = globalpolicy.getMaxFailedLogins()
        if globalpolicy.getPolicyDisabled() not in (None, ""):
            # policy disabled globally, so clear retry count and lock_after_retries
            changed = self.setLockAfterRetries(result, "NO", apply) or changed
            changed = self.setRetries(result, None, apply) or changed
        elif gf is not None and int(gf) > 0:
            # global maxFailedLogins is nonzero: set retry count, and enable lock_after_retries.
            changed = self.setLockAfterRetries(result, "YES", apply) or changed
            changed = self.setRetries(result, gf, apply) or changed
        elif minUserFail[0] > 0:
            # global maxFailedLogins is unset or zero, and one or more users have non-zero values
            # set the retry count to the minimum of those, and disable lock_after_retries.
            changed = self.setLockAfterRetries(result, "NO", apply) or changed
            changed = self.setRetries(result, minUserFail[0], apply) or changed
        else:
            # global maxFailedLogins is unset or zero and no users have nonzero values
            # clear the retry count and disable lock_after_retries.
            changed = self.setLockAfterRetries(result, "NO", apply) or changed
            changed = self.setRetries(result, None, apply) or changed

        result.finishAction(npm_s_success, [])
        return changed


    def policyUpdate(self, result, delta, apply):
        gp = self.getGlobalPolicy(delta.newPolicy)
        up = self.getUserPolicy(delta.newPolicy)

        # if there's any applicable policy at all before and after this delta,
        # then we're not going to manage the login activity policy at all.
        if gp is None and up == {}:
            ogp = self.getGlobalPolicy(delta.oldPolicy)
            oup = self.getUserPolicy(delta.oldPolicy)
            if ogp is None and oup == {}:
                # no policy, do nothing
                result.finish(npm_s_success, [])
                return

        if gp is None:
            gp = LoginActivity("default", "TAMOS")

        changed = self.applyLoginPolicy(result, gp, up, apply)
        if apply == True or changed == False:
            result.finish(npm_s_success, [])
        else:
            result.finish(npm_s_policy_not_compliant, [])


    def verifyPolicy(self, result, policy, reconcile):

        gp = self.getGlobalPolicy(policy)
        up = self.getUserPolicy(policy)
        
        if gp is None and up == {}:
            # no policy, so do nothing
            result.finish(npm_s_success, [])
            return

        if gp is None:
            gp = LoginActivity("default", "TAMOS")

        changed = self.applyLoginPolicy(result, gp, up, reconcile)
        if changed is False:
            result.finish(npm_s_success, [])
        elif reconcile:
            result.finish(npm_s_policy_reconciled, [])
        else:
            result.finish(npm_s_policy_not_compliant, [])

# vim: set ts=4 sw=4 expandtab :
