#
# IBM Confidential
# OCO Source Materials
# Tivoli Access Manager for Operating Systems - Agent
# TIVO-TAMOSA
#
# (C) Copyright IBM Corp. 2008
# The source code for this program is not published or otherwise
# divested of its trade secrets, irrespective of what has been
# deposited with the U.S. Copyright Office.
#

#
# common code for dealing with AIX auditing
#

import java.lang.StringBuffer
import re
from com.ibm.amos.agent.db import PolicyConstants
from com.ibm.amos.agent.db.rules import Audit
from com.ibm.amos.agent.db.resources import File

# Audit configuration files
AUDIT_OBJECTS = "/etc/security/audit/objects"

def restartAuditDaemon(target, result):
   """
   Restart the audit daemon.  We use a helper script for this because
   we need to run through su (to reset group memberships) and
   close stdout and stderr when running 'audit start', neither of which
   we can do directly.
   """
   out = java.lang.StringBuffer()
   err = java.lang.StringBuffer()
   result.newStep("Restarting audit system")
   rc = target.execute(("/opt/pdos/sbin/pdosauditrestart",), None, out, err)
   if rc != 0:
      target.logger.finer("An error occurred restarting the audit system.")
      target.logger.finer(out.toString())
      target.logger.finer(err.toString())

# Policy mappings
audit_mapping = {
                  "r":["r"],
                  "w":["w"],
                  "N":["w"],
                  "x":["x"],
                  "o":["w"],
                  "p":["w"],
                  "D":["x"],
                  "R":["w"],
                  "d":["w"],
                  "U":["w"],
                  "l":["r","x"]
                }

action_defaults = {
                     "r":"FILE_Open",
                     "w":"FILE_Open",
                     "x":"FILE_Open"
                  }

action_attributes = {
                     "r":"aix_audit_read_event",
                     "w":"aix_audit_write_event",
                     "x":"aix_audit_execute_event"
                    }

audit_actions = { "permit" : "audit_permit_actions",
                  "deny"   : "audit_deny_actions" }


valid_audit_levels = [ "permit", "deny" ]

def getActionEntry(flag, auditObj):
   """
   Retrieve the AIX audit class associated with the flag.
   The flag is one of 'rwx', the action entry is either the value of
   the appropriate aix_audit_<type>_event set on the audit object or
   the default as set in action_defaults
   """

   entry = None
   if auditObj.getExtAttributes() != None and action_attributes[flag] in auditObj.getExtAttributes():
      entry = auditObj.getExtAttributes().get(action_attributes[flag]).get(0)
   else:
      entry = action_defaults[flag]
   return entry


def getActionEntries(auditObj):
   """
   Methods for creating the action entries dictionary for the audit objects
   class.
   This is a mapping from r, w or x to the AIX audit class that should be
   set for that access type.
   """

   action_entries = {}
   flags = ""

   levels = auditObj.getLevels()
   if levels is None:
      return action_entries

   for level in levels:
      if level in audit_actions:
         audit_action = audit_actions[level]
         ea = auditObj.getExtAttributes()
         if ea is not None and audit_action in ea:
            flagsre = re.search("\[OSSEAL\]([a-zA-Z]*)", ea.get(audit_actions[level]).get(0))
            flags = flags + flagsre.group(1)
         else:
            # "rwx" is a representative sample
            flags = flags + "rwx"
            
   for flag in flags:
      if flag in audit_mapping:
         events = audit_mapping[flag]
         for event in events:
            if event not in action_entries:
               action_entries[event] = getActionEntry(event, auditObj)

   return action_entries


def isValidAudit(auditID, policy):
   """
   Check if there is a valid audit object with the ID specified in the
   policy. A valid audit object is one that exists and has at least one
   level from the valid_audit_levels (ie. 'permit' or 'deny')
   """
   if auditID is None or policy is None:
      return False
   
   auditmap = policy.ruleMap.get(PolicyConstants.AUDIT)
   
   if auditmap is None:
      return False
   
   auditObj = auditmap.get(auditID)
   if auditObj is not None:
      for lvl in valid_audit_levels:
         if lvl in auditObj.getLevels():
            return True
         
   return False


def isValidAuditOnFile(filename, policy):
   """
   Check if there is a valid audit object attached to the file specified
   in the policy.
   """

   if filename is None or policy is None:
      return False
   
   filemap = policy.resourceMap.get(PolicyConstants.FILE)
   auditmap = policy.ruleMap.get(PolicyConstants.AUDIT)
   if filemap is None or auditmap is None:
      return False
   
   file = filemap.get(filename)
   if file is not None:
      auditID = file.getAuditID()
      return isValidAudit(auditID, policy)

   return False


