#
# IBM Confidential
# OCO Source Materials
# Tivoli Access Manager for Operating Systems - Agent
# TIVO-TAMOSA
#
# (C) Copyright IBM Corp. 2008
# The source code for this program is not published or otherwise
# divested of its trade secrets, irrespective of what has been
# deposited with the U.S. Copyright Office.
#

#
# configuration file editing utilities
#

from __future__ import generators

class LineEditor:
    """
    Line-based file editor.
    """

    def __init__(self, data, has_comments = True, has_continuations = False):
        self.data = str(data)
        self.comments = has_comments
        self.continuations = has_continuations
        self.changed = False

    def hasChanged(self):
        return self.changed

    def getData(self):
        return self.data

    def rawlines(self):
        """
        Generator returning raw lines from the file, ignoring comments
        and continuations etc.
        """
        linestart = 0
        idx = 0
        while idx < len(self.data):
            if self.data[idx] == '\n':
                yield self.data[linestart:idx]
                linestart = idx+1
            idx = idx+1

        yield self.data[linestart:]

    def lines(self):
        """
        Generator yielding lines from the file, accounting for comments and
        continuations.
        """
        line = ""
        for l in self.rawlines():
            sl = l.strip()
            if sl == "" or (self.comments and sl.startswith('#')):
                continue

            line += sl
            if self.continuations and sl.endswith('\\'):
                line = line[:-1]
                continue

            yield line
            line = ""

        if line != "":
            yield line

    def append(self, line):
        # do some ugly tricks for trailing blank lines
        if self.data not in (None, "") and self.data[-1] != "\n":
            self.data += "\n"
        self.data += line
        if self.data[-1] != "\n":
            self.data += "\n"
        self.changed = True

    def update(self, callback, data):
        """
        Calls the callback for each line in turn.
        The callback returns the new content of the line, or 
        None if the line should be deleted.  Comments are preserved.
        """

        newdata = ""

        line = ""
        origline = ""
        for l in self.rawlines():
            if line == "":      # don't consider comments when processing continuations
                sl = l.lstrip()
                if sl == "" or (self.comments and sl.startswith('#')):
                    newdata += "\n" + l
                    continue

            line += l
            origline += l
            if self.continuations and l.endswith('\\'):
                line = line[:-1]
                continue

            newline = callback(line, data)
            if newline == line:
                # preserve original formatting
                newdata += "\n" + origline
            elif newline is not None:
                self.changed = True
                newdata += "\n" + newline
            else:
                self.changed = True

            line = ""
            origline = ""

        if line != "":
            newline = callback(line, data)
            if newline == line:
                # don't preserve broken formatting in this case
                newdata += "\n" + line
                if line != origline:
                    self.changed = True
            elif newline is not None:
                self.changed = True
                newdata += "\n" + newline
            else:
                self.changed = True

        self.data = newdata[1:]     # skip leading newline


class Stanza:
    def __init__(self, name, lines, coda = None):
        self.name = name
        self.lines = lines
        self.changed = False
        self.coda = coda

    def parse(self):
        if self.coda is not None and len(self.lines) > 0:
            if self.lines[-1] == self.coda:
                self.lines = self.lines[:-1]
    
    def append(self, line):
        """
        Add a line to the end of a stanza
        """
        self.lines.append(line)

    def getText(self):
        """
        Get the text of the stanza. 
        """
        stzText = ""
        if len(self.lines) > 0:
            stzText = "\n".join(self.lines)

            if self.coda is not None:
                stzText += "\n%s" % self.coda

        return stzText
    
    def getName(self):
        """
        Get the stanza name
        """
        return self.name

    def update(self, callback, cbdata):
        """
        Update the existing stanza entries by running a call back
        procedure over all the lines. If None is returned from the callback
        the line is removed, otherwise the new value is used.
        """
        def cb(item):
            return callback(self.name, item, cbdata)

        newlines = map(cb, self.lines)
        nonEmptyNewlines = [l for l in newlines if l is not None]
        if not self.lines == nonEmptyNewlines:
            self.changed = True
            self.lines = nonEmptyNewlines

    def hasChanged(self):
        return self.changed

    def getEntryCount(self):
        return max(0, len(self.lines) - 1)

class StanzaFile:
    def __init__(self, data, has_comments = True, has_continuations = True, stanza_coda = None):
        self.edit = LineEditor(data, has_comments, has_continuations)
        self.changed = False
        self.stzCoda = stanza_coda


    def hasChanged(self):
        """
        Has the file changed?
        """
        return self.changed

    def determineStanzaName(self, line):
        """
        A function to determine if the line supplied is indicates that a new
        stanza has begun. This should return either the new stanza name or
        None.
        """
        return None

    def determineStanzaLine(self, name):
        """
        A function to determine the line that should be added for the
        stanza supplied. 
        """
        return name
        

    def rawStanza(self):
        """
        A generator for retrieving stanzas from the data
        """
        stanza = None

        for line in self.edit.rawlines():
            nextStanzaName = self.determineStanzaName(line)
            if nextStanzaName is not None:
                if stanza is not None:
                    yield stanza
                stanza = Stanza(nextStanzaName, [line], self.stzCoda)
            else:
                if stanza is None:
                    stanza = Stanza(None, [], self.stzCoda)
                stanza.append(line)
        if stanza is not None:
            yield stanza

    def parse(self):
        """
        Parse the data supplied into stanzas so that it can be manipulated
        later
        """
        self.stanzas = []
        self.stanzaMap = {}
        for stanza in self.rawStanza():
            self.stanzas.append(stanza)
            self.stanzaMap[stanza.getName()] = stanza
            stanza.parse()

    def addStanza(self, stanza):
        """
        Add a new stanza to the file
        """
        self.stanzas.append(stanza)
        self.stanzaMap[stanza.getName()] = stanza

    def getText(self):
        """
        Retrieve the text of the file, by retrieving the text of all the
        stanzas in the file
        """
        def cb(item):
            return item.getText()

        return "\n".join(map(cb, self.stanzas))

    def append(self, stanzaName, line):
        """
        Append a line to a named stanza. If the stanza doesn't exist, a
        stanza by that name is 
        """
        if self.stanzaMap.has_key(stanzaName):
            self.stanzaMap[stanzaName].append(line)
        else:
            stanza = Stanza(stanzaName, [self.determineStanzaLine(stanzaName), line], self.stzCoda)
            self.addStanza(stanza)

        self.changed = True

    def updateStz(self, stanza, callback, cbdata):
        stanza.update(callback, cbdata)
        if stanza.hasChanged():
            self.changed = True

    def update(self, stanzaName, callback, cbdata):
        """
        Update the entries in a particular stanza, based on a callback
        routine.
        """
        if self.stanzaMap.has_key(stanzaName):
            stanza = self.stanzaMap[stanzaName]
            self.updateStz(stanza, callback, cbdata)

    def updateAllStanzas(self, callback, cbdata):
        """
        Perform an update across all the stanzas in the file, based on a
        callback routine.
        """
        for stanza in self.stanzas:
            self.updateStz(stanza, callback, cbdata)

    def getStanzas(self):
        return self.stanzas

    def getStanza(self, stanzaName):
        if self.stanzaMap.has_key(stanzaName):
            return self.stanzaMap[stanzaName]
        return None

    def getStanzaEntryCount(self, stanzaName):
        if self.stanzaMap.has_key(stanzaName):
            return self.stanzaMap[stanzaName].getEntryCount()
        return None



# vim: set ts=4 sw=4 expandtab :
