#
# IBM Confidential
# OCO Source Materials
# Tivoli Access Manager for Operating Systems - Agent
# TIVO-TAMOSA
#
# (C) Copyright IBM Corp. 2008
# The source code for this program is not published or otherwise
# divested of its trade secrets, irrespective of what has been
# deposited with the U.S. Copyright Office.
#

#
# common code for dealing with Solaris auditing
#

# audit configuration files
AUDIT_USER = "/etc/security/audit_user"
AUDIT_CONTROL = "/etc/security/audit_control"
AUDIT_CLASSES = "/etc/security/audit_classes"

# TAMOS audit attributes
PERMIT_ACTIONS = "audit_permit_actions"
DENY_ACTIONS = "audit_deny_actions"

# default solaris audit metaclasses
# we should probably read these from /etc/security/audit_classes at runtime .. but not yet.
audit_metaclass_map = {
       'am': ('ss', 'as', 'ua'),
       'ad': ('aa', 'ua', 'as', 'ss'),
       'pc': ('ps', 'pm'),
       'all': ('fr', 'rw', 'fa', 'fm', 'fc', 'fd', 'cl', 'nt', 'ip', 'na', 'lo', 'ap', 'ss', 'as', 'ua', 'aa', 'ps', 'pm', 'xp', 'xc', 'xs', 'io', 'ex', 'ot')
}


def parseAuditFlag(flag):
        """
        Given a Solaris audit flag, returns a tuple containing the audit
        class (0), the sense in which the flag applies (1), and whether it applies
        to permit (2) and deny (3) events.
        """
        auditclass = flag
        (sense, set_permit, set_deny) = (True, True, True)
        
        if auditclass[0] == '^':
                sense = False
                auditclass = auditclass[1:]
        
        if auditclass[0] == '+':
                set_deny = False
                auditclass = auditclass[1:]
        elif auditclass[0] == '-':
                set_permit = False
                auditclass = auditclass[1:]
        
        return (auditclass, sense, set_permit, set_deny)



def parseAndExpandAuditFlag(flag, meta):
        """
        Parses an audit flag (as per parseAuditFlag) and expands metaclasses from
        the given metaclass map
        """
        (klass, sense, set_permit, set_deny) = parseAuditFlag(flag)
        if klass in meta:
                classes = meta[klass]
        else:
                classes = (klass,)
        
        return (classes, sense, set_permit, set_deny)



def evaluateAuditClass(flags, auditclass, meta):
        """
        Determines whether an audit flag string enables auditing for permit and deny events
        for a particular class.  The metaclass map provided must map metaclass names to
        lists of classes that make up the metaclass.
        
        Returns a tuple describing whether auditing for permit (0) and deny (1) events
        is enabled for the class.
        """
        deny = False
        permit = False
        for f in flags.split(","):
                if f == "":
                        continue
                
                # split out modifiers
                (classes, sense, set_permit, set_deny) = parseAndExpandAuditFlag(f, meta)
                
                # does this flag apply to the class we're interested in?        
                if auditclass in classes:
                        if set_permit:
                                permit = sense
                        if set_deny:
                                deny = sense

        return (permit, deny)



def evaluateAuditFlagString(flags, meta):
	"""
	Parses an audit flag string, returning a tuple containing (0) the classes enabled
	for both permit and deny, (1) the classes enabled for permit only, and (2) the
	classes enabled for deny only.  Class sets are returned as dicts with class names
	as keys and nothing important as values.  Python 2.4+ has built-in set objects
	that would be better to use here.
	"""
        both_classes = {}
        permit_classes = {}
        deny_classes = {}
        
        # evaluate the string, identifying classes it enables for permit and deny
        for f in flags.split(","):
                if f == "":
                        continue
                
                (classes, sense, set_permit, set_deny) = parseAndExpandAuditFlag(f, meta)
                for klass in classes:
                        def __apply(classes, klass, sense):
                                if sense:
                                        classes[klass] = 1
                                elif klass in classes:
                                        del classes[klass]
                                        
                        if set_permit:
                                __apply(permit_classes, klass, sense)
                        if set_deny:
                                __apply(deny_classes, klass, sense)

        # extract classes that are enabled for both permit and deny
        for klass in permit_classes.iterkeys():
                if klass in deny_classes:
                        both_classes[klass] = 1
        
        for klass in both_classes:
                del permit_classes[klass]
                del deny_classes[klass]

	return (both_classes, permit_classes, deny_classes)



def compareFlags(a, b, meta):
	"""
	Returns True if the two audit flag strings are equivalent.
	"""
	eval_a = evaluateAuditFlagString(a, meta)
	eval_b = evaluateAuditFlagString(b, meta)
	return eval_a == eval_b



def normalizeAuditFlags(flags, meta):
        """
        Collapses an audit flag string into (probably) the shortest equivalent form.
        This is mostly here so we can enable audit classes by appending ",<class>"
        or remove them by appending ",^<class>" without having to think too hard.
        """
        
	(both_classes, permit_classes, deny_classes) = evaluateAuditFlagString(flags, meta)

        def collapseMetaClasses(classlist, meta):
                for m in meta.iterkeys():
                        for mem in meta[m]:
                                if mem not in classlist:
                                        break
                        else:
                                # all members of metaclass present, so collapse it
                                for mem in meta[m]:
                                        del classlist[mem]
                                classlist[m] = []
        
        # collapse metaclasses
        # we could probably apply metaclasses where a small
        # number of exclusions were required afterwards, but
        # it's not really seeming like it's worth it right now.
        collapseMetaClasses(both_classes, meta)
        collapseMetaClasses(permit_classes, meta)
        collapseMetaClasses(deny_classes, meta)

        def addFlags(newflags, classes, modifier):
                k = classes.keys()
                k.sort()
                for c in k:
                        newflags.append(modifier + c)
        
        newflags = []
        addFlags(newflags, both_classes, "")
        addFlags(newflags, permit_classes, "+")
        addFlags(newflags, deny_classes, "-")
        return ",".join(newflags)

