#
# IBM Confidential
# OCO Source Materials
# Tivoli Access Manager for Operating Systems - Agent
# TIVO-TAMOSA
#
# (C) Copyright IBM Corp. 2008
# The source code for this program is not published or otherwise
# divested of its trade secrets, irrespective of what has been
# deposited with the U.S. Copyright Office.
#

# various wildcard related methods

from __future__ import generators

from ossutil import splitDirName
import re

# These constants are ordered in priority from highest priority to lowest
# priority. This is taken from the TAMOS 6.0 Admin Guide, Chapter 2, Page 12

EXACT_CHARACTER = 1
CHARACTER_RANGE = 2
ARBITRARY_CHARACTER = 3
REPEATED_EXACT_CHARACTERS = 4
REPEATED_CHARACTER_RANGE = 5
REPEATED_ARBITRARY_CHARACTER = 6
ARBITRARY_STRING = 7
INVALID_REGEXP = 8

def escape(filename):
    """
    Returns a copy of a filename with wildcard metacharacters escaped
    """
    esc = ""
    i = 0
    n = len(filename)
    while i < n:
        c = filename[i]
        i = i+1

        if c in ('*', '?', '[', ']', '+', '\\'):
            esc = esc + "\\"
        esc = esc + c

    return esc


def convertToRegex(pattern):
    """
    Converts a wildcard specification to a regular expression or an
    unescaped form.  what?
    """
    # scan the string, producing an unescaped version of the filename
    # and a regular expression form of the pattern.  if the filename contains
    # any wildcard metacharacters, the unescaped version of the filename is
    # useless and broken.
    unesc = ""
    regex = ""
    iswc = False
    i = 0
    n = len(pattern)
    while i < n:
        c = pattern[i]
        i = i+1

        if c == '*':
            iswc = True
            regex = regex + ".*"
        elif c == '?':
            iswc = True
            regex = regex + "."
        elif c == '+':
            iswc = True
            regex = regex + "+"
        elif c == '[':
            iswc = True
            j = i
            while j < n and pattern[j] != ']':
                j = j+1

            if j >= n:
                regex = regex + "\\["
            else:
                regex = regex + "[" + pattern[i:j] + "]"
                i = j+1

        elif c == '\\':
            if i < n:
                unesc = unesc + pattern[i]
                regex = regex + re.escape(pattern[i])
                i = i+1
        else:
            unesc = unesc + c
            regex = regex + re.escape(c)

    if iswc:
        return (True, regex + "$")
    else:
        return (False, unesc)


def extract(filename):
    """
    If the filename is a wildcard specification, returns (dir, pattern, regex);
    otherwise, if the directory contains any wildcard specifications, returns (None, None, None);
    otherwise, returns (dir, unescaped, None).
    """

    (dir, fn) = splitDirName(filename)

    (dirwc, unesc_dir) = convertToRegex(dir)
    if dirwc:
        # directory contains wildcards
        return (None, None, None)

    (filewc, f) = convertToRegex(fn)
    if filewc:
        return (unesc_dir, fn, f)
    else:
        return (unesc_dir, f, None)



def get_element_type(element):
    """
    Determine which element class the identified element belongs to
    """
    elelen = len(element)

    if element == "*":
        return ARBITRARY_STRING
    elif element == "?":
        return ARBITRARY_CHARACTER
    elif element == "?+":
        return REPEATED_ARBITRARY_CHARACTER
    elif elelen == 1:
        return EXACT_CHARACTER
    else:
        if elelen == 2 and element[0] == '\\':
            return EXACT_CHARACTER
        elif element[elelen - 1] == '+':
            if element[0] == '?':
                return REPEATED_ARBITRARY_CHARACTER
            elif elelen == 2 or (elelen == 3 and element[0] == '\\'):
                return REPEATED_EXACT_CHARACTERS
            elif element[0] == '[' and element[elelen - 2] == ']':
                return REPEATED_CHARACTER_RANGE
            else:
                return INVALID_ELEMENT
        elif element[0] == '[' and element[elelen - 1] == ']':
            return CHARACTER_RANGE
        else:
            return INVALID_ELEMENT



def get_element(thestr):
    """
    Generator to retrieve the next regular expression element from the
    regular expression.
    """
    start = 0
    end = 0

    in_range = 0
    is_escaped = False
    peeked_plus = False

    for char in thestr:
        end = end + 1
        
        if peeked_plus:
            yield thestr[start:end]
            peeked_plus = False
            start = end
        else:
            if is_escaped:
                is_escaped = False
                if in_range == 0:
                    if end < len(thestr) and thestr[end] == '+':
                        peeked_plus = True
                    else:
                        yield thestr[start:end]
                        start = end
            else:
                if in_range > 0:
                    if char == '[':
                        in_range = in_range + 1
                    elif char == ']':
                        in_range = in_range - 1
                    elif char == '\\':
                        is_escaped = True

                    if in_range == 0:
                        if end < len(thestr) and thestr[end] == '+':
                            peeked_plus = True
                        else:
                            yield thestr[start:end]
                            start = end

                elif char == '\\':
                    is_escaped = True

                elif char == '[':
                    in_range = in_range + 1

                elif char == '*':
                    # Asterisk can't be followed by '+', so don't check.
                    yield thestr[start:end]
                    start = end
                else:
                    # Normal character
                    if end < len(thestr) and thestr[end] == '+':
                        peeked_plus = True
                    else:
                        yield thestr[start:end]
                        start = end
            

    if start < len(thestr):                        
        yield thestr[start:end]


def compare_regexp(reg_a, reg_b):
    """
    Comparator for regular expressions. This orders regular expressions from
    highest priority to lowest priority.
    
    """
    gen_a = get_element(reg_a)
    gen_b = get_element(reg_b)
    
    elt_a = gen_a.next()
    elt_b = gen_b.next()

    while elt_a is not None and elt_b is not None:
        type_a = get_element_type(elt_a)
        type_b = get_element_type(elt_b)
        if not type_a == type_b:
            return type_a - type_b

        try:
            elt_a = gen_a.next()
        except StopIteration:
            elt_a = None

        try:
            elt_b = gen_b.next()
        except StopIteration:
            elt_b = None


    if elt_a is None and elt_b is None:
        return 0

    if elt_a is not None and elt_b is None:
        type_a = get_element_type(elt_a)
        if not type_a == ARBITRARY_STRING:
            return -1
        else:
            try:
                next_a = gen_a.next()
            except StopIteration:
                next_a = None

            if next_a is None:
                return 1
            else:
                return -1
    elif elt_b is not None and  elt_a is None:
        type_b = get_element_type(elt_b)
        if not type_b == ARBITRARY_STRING:
            return 1
        else:
            try:
                next_b = gen_b.next()
            except StopIteration:
                next_b = None

            if next_b is None:
                return -1
            else:
                return 1
    else:
        assert False



# vim: set ts=4 sw=4 expandtab :
