#
# IBM Confidential
# OCO Source Materials
# Tivoli Access Manager for Operating Systems - Agent
# TIVO-TAMOSA
#
# (C) Copyright IBM Corp. 2008
# The source code for this program is not published or otherwise
# divested of its trade secrets, irrespective of what has been
# deposited with the U.S. Copyright Office.
#

#
# aix password policy implementation
# 

from com.ibm.amos.agent.db import PolicyConstants
from com.ibm.amos.agent.db.resources import User
from com.ibm.amos.agent.db.resources import File
from com.ibm.amos.agent.db.rules import PasswordManagement
from com.ibm.amos.nls.pdopmidx import *

import java.lang.Exception
import java.lang.String
import java.lang.StringBuffer
import java.util.logging.Level

from ossutil import wrapPyException
from ossutil.configeditor import StanzaFile
from ossutil.aixconfig import AixSystemConfigFile

PASSWORD_POLICY_CONFIG_FILE = "/etc/security/user"

DEFAULT_USER = "default"

MIN_LEN = "minlen"
MIN_ALPHA = "minalpha"
MIN_OTHER = "minother"
MIN_AGE = "minage"
MAX_REPEATS = "maxrepeats"
MIN_DIFF = "mindiff"
HIST_SIZE = "histsize"

CONTROLLED_VALUES = ( MIN_LEN, MIN_ALPHA, MIN_OTHER, MIN_AGE, MAX_REPEATS, MIN_DIFF, HIST_SIZE )

class AixPasswordPolicyPlugin:

    def __init__(self, realplugin, target):
        self.target = target
        self.realplugin = realplugin
        self.__l = target.logger

    def shutdown(self):
        # do nothing
        pass

    def applyPolicy(self, result, policy, update, warnOnly = False):
        """
        Wrapper around the apply policy implementation that loads / stores the
        configuration file, mainly for ease of unit testing.
        """
        outcome = npm_s_success

        try:
            # Read the current state
            xau = self.target.translatePath(PASSWORD_POLICY_CONFIG_FILE)
            data = java.lang.String(self.target.getFile(xau))
            passwordPolicyFile = AixSystemConfigFile(data)

            outcome = self.applyPolicyImpl(result, passwordPolicyFile, policy, update, warnOnly)

            if passwordPolicyFile.hasChanged():
                result.startAction(File(PASSWORD_POLICY_CONFIG_FILE, None, None))
                result.newStep("Updating password policy file")
                self.__l.finest("New config file content:\n%s\n" % (passwordPolicyFile.getText()))
                self.target.putFile(xau, passwordPolicyFile.getText(), True)
		result.finishAction(npm_s_success, [])

            result.finish(outcome, [])

        except java.lang.Exception, e:
            result.finish(e)
        except Exception, e:
            result.finish(wrapPyException(self.realplugin, e))


    def getGlobalPasswordPolicy(self, target, policy):
        """
        Retrieves the global password policy. An empty global password policy
        is considered the same as a non-present global password policy.
        """
        global_pwdpol = None
        global_res = policy.resourceMap[PolicyConstants.GLOBAL]
        if global_res is not None:
            if len(global_res) == 1:
                for name in global_res:
                    elt = global_res[name]
                    global_pwdpol_id = elt.getPasswdID()
                    if global_pwdpol_id is not None:
                        global_pwdpol = policy.getRule(PolicyConstants.PASSWORD_MANAGEMENT, global_pwdpol_id)
            else:
                self.__l.finer("Unexpected number of global elements found.")

        # Check whether or not there's an attribute set. We don't particularly
        # care which one
        if global_pwdpol is not None:
            empty_pol = PasswordManagement("empty", "TAMOS")
            if empty_pol.equals(global_pwdpol):
                global_pwdpol = None

        return global_pwdpol


    def determinePasswordPolicyUsers(self, target, policy):
        """
        Figures out which password policy settings must be applied to which users
        """
        users = {}

        def processUser(users, name, pwdpol, min_len_fallback, maxconsprev_fallback):
            # TAMOS specifies min age value in days, AIX specifies them in weeks
            # We round the value down to a minimum of 1 for non-zero values.
            def age_conversion(value_str):
                if value_str is not None:
                    value = int(value_str)
                    if value < 1:
                        value = 0
                    else:
                        value = max(1, value / 7)
                    return value
                return None

            def make_non_negative(value_str):
                if value_str is not None:
                    return max(0, int(value_str))
                return None

            def diff_conversion(pwd_len_str, min_len_fallback, max_cons_prev_str, maxconsprev_fallback):
                if pwd_len_str is not None or max_cons_prev_str is not None:
                    min_len = None
                    max_consprev = None

                    if pwd_len_str is not None:
                        min_len = make_non_negative(pwd_len_str)
                    else:
                        min_len = min_len_fallback

                    if max_cons_prev_str is not None:
                        max_consprev = make_non_negative(max_cons_prev_str)
                    else:
                        max_consprev = maxconsprev_fallback

                    if min_len is not None and max_consprev is not None:
                        if max_consprev > 0:
                            return max(0, min_len - max_consprev)
                        else:
                            return 0
                    
                return None

            currUserValues = {}
            currUserValues[MIN_LEN]     = make_non_negative(pwdpol.getMinPasswordLength())
            currUserValues[MIN_ALPHA]   = make_non_negative(pwdpol.getMinPasswordAlpha())
            currUserValues[MIN_OTHER]   = make_non_negative(pwdpol.getMinPasswordSpecial())
            currUserValues[MIN_AGE]     = age_conversion(pwdpol.getMinPasswordDays())
            currUserValues[MAX_REPEATS] = make_non_negative(pwdpol.getMaxPasswordRepeat())
            currUserValues[MIN_DIFF]    = diff_conversion(pwdpol.getMinPasswordLength(), min_len_fallback, pwdpol.getMaxConsecutivePrevious(), maxconsprev_fallback)
            currUserValues[HIST_SIZE]   = make_non_negative(pwdpol.getPasswordHistorySize())

            haveValue = 0
            for elt in CONTROLLED_VALUES:
                if currUserValues[elt] is not None:
                    haveValue = 1
                    break

            # If they have a password policy value set, copy across the values that are set
            if haveValue == 1:
                if not users.has_key(name):
                    users[name] = {}
                for elt in CONTROLLED_VALUES:
                    if currUserValues[elt] is not None:
                        users[name][elt] = currUserValues[elt]


        
        # Find the global password policy. We do this first so that we know the
        # default value for minPasswordLength for use with determining the
        # MIN_DIFF value
        global_minlen = None
        global_maxconsprev = None
        global_pwdpol = self.getGlobalPasswordPolicy(target, policy)
        users[DEFAULT_USER] = {}
        for value in CONTROLLED_VALUES:
            users[DEFAULT_USER][value] = 0

        if global_pwdpol is not None:
            processUser(users, DEFAULT_USER, global_pwdpol, None, None)
            if global_pwdpol.getMinPasswordLength() is not None:
                global_minlen = max(0, int(global_pwdpol.getMinPasswordLength()))
            if global_pwdpol.getMaxConsecutivePrevious() is not None:
                global_maxconsprev = max(0, int(global_pwdpol.getMaxConsecutivePrevious()))

        # Find the per-user password policy
        userRes = policy.resourceMap[PolicyConstants.USER]
        if userRes is not None:
            for name in userRes:
                if name == DEFAULT_USER:
                    # Log a warning about this
                    pass
                else:
                    user = userRes[name]
                    user_pwdpol_id = user.getPasswdID()
                    if user_pwdpol_id is not None:
                        user_pwdpol = policy.getRule(PolicyConstants.PASSWORD_MANAGEMENT, user_pwdpol_id)
                        if user_pwdpol is not None:
                            processUser(users, name, user_pwdpol, global_minlen, global_maxconsprev)

        # If we don't have a global password policy and only have the default
        # entry in users, then we really have no password policy set at all
        if global_pwdpol is None and len(users) == 1:
            users = {}

        return users


    def applyPolicyImpl(self, result, pwdFile, policy, update, warnOnly):
        """
        Implementation of actual policy application. This has been implemented
        as a separte function mainly for ease of unit testing.
        """
        outcome = npm_s_success

        pwdFile.parse()

        users = self.determinePasswordPolicyUsers(self.target, policy)

        def updateCb(stanzaName, name, values, data):
            (result, users) = data
            result.startAction(User(name, "TAMOS"))

            if name in CONTROLLED_VALUES:
                retval = None
            else:
                retval = values

            if users.has_key(stanzaName):
                if users[stanzaName].has_key(name):
                    retval = ["%d" % (users[stanzaName][name])]
                    del users[stanzaName][name]
                    if len(users[stanzaName].keys()) == 0:
                        del users[stanzaName]

            if not retval == values:
                if retval is None:
                    result.diff("user", name, ",".join(values), None)
                else:
                    result.diff("user", name, ",".join(values), ",".join(retval))

            if update or retval == values:
                result.finishAction(npm_s_success, [])
            elif warnOnly:
                result.finishAction(npm_s_policy_not_compliant, [])
                retval = values
            else:
                result.finishAction(npm_s_policy_reconciled, [])

            return retval

        # If there is no password policy set, we assume this means that they
        # are not using TAMOS to centrally manage password policy, so we leave
        # any existing policy unmodified.
        if len(users.keys()) > 0:
            self.__l.finer("Updating any existing password management policy settings.")
            pwdFile.updateAllStanzas(updateCb, (result, users))
            self.__l.finer("Adding any remaining entries with no prior login activity policy setting")

            for user in users:
                for elt in users[user]:
                    result.startAction(User(user, "TAMOS"))
                    values = ["%d" % (users[user][elt])]
                    result.diff("user", user, None, ",".join(values))
                    if update:
                        pwdFile.appendEntry(user, elt, values)
                        result.finishAction(npm_s_success, [])
                    else:
                        if warnOnly:
                            result.finishAction(npm_s_policy_not_compliant, [])
                        else:
                            pwdFile.appendEntry(user, elt, values)
                            result.finishAction(npm_s_policy_reconciled, [])
        else:
            self.__l.finer("No password management policy has been defined, leaving existing configuration unchanged.")

        return outcome


    def policyUpdate(self, result, delta, apply):
        """
        Apply a policy update to the user audit settings.
        """
        self.applyPolicy(result, delta.newPolicy, True, False)
        

    def verifyPolicy(self, result, policy, reconcile):
        """
        Verify that the policy that is currently on the system complies with
        what is set in TAMOS. Optionally, update the policy to reflect with
        what TAMOS has.
        """
        self.applyPolicy(result, policy, False, not reconcile)
        
        

    
