#
# IBM Confidential
# OCO Source Materials
# Tivoli Access Manager for Operating Systems - Agent
# TIVO-TAMOSA
#
# (C) Copyright IBM Corp. 2008
# The source code for this program is not published or otherwise
# divested of its trade secrets, irrespective of what has been
# deposited with the U.S. Copyright Office.
#

#
# solaris password policy implementation
#

from com.ibm.amos.agent.db import PolicyConstants
from com.ibm.amos.agent.db.rules import PasswordManagement
from com.ibm.amos.agent.db.resources import Global, User
from com.ibm.amos.nls.pdopmidx import *

from ossutil.keyvalueeditor import KeyValueEditor
from ossutil.solaris import SolarisShadowFile

import java.lang.String

SHADOW = "/etc/shadow"
PASSWD_DEFAULTS = "/etc/default/passwd"


def getEnablePasswordNameCheck(policy):
    v = policy.getEnablePasswordNameCheck()
    if v not in (None, "") and int(v) > 0:
        return "yes"
    return "no"

def getMinDiff(policy):
    # mindiff = max(0, minPasswordLength - maxConsecutivePrevious)
    maxconsec = policy.getMaxConsecutivePrevious()
    len = policy.getMinPasswordLength()
    if maxconsec in (None, "") or len in (None, ""):
        # we need both these fields defined to have a value here
        return None

    return str(max(0, int(len) - int(maxconsec)))

def getPasswordHistorySize(policy):
    # maximum history size is 26
    v = policy.getPasswordHistorySize()
    if v in (None, ""):
        return None

    return str(max(0, min(int(v), 26)))


mapping = {
    # method to retrieve value: ( field, default )
    PasswordManagement.getMinPasswordLength: ( 'PASSLENGTH', '0' ),
    PasswordManagement.getMinPasswordAlpha: ( 'MINALPHA', '0' ),
    PasswordManagement.getMinPasswordNumeric: ( 'MINDIGIT', None ),
    PasswordManagement.getMinPasswordLower: ( 'MINLOWER', None ),
    PasswordManagement.getMinPasswordUpper: ( 'MINUPPER', None ),
    PasswordManagement.getMinPasswordSpecial: ( 'MINSPECIAL', None ),
    PasswordManagement.getMaxPasswordRepeat: ( 'MAXREPEATS', None ),
    getPasswordHistorySize: ( 'HISTORY', None ),
    getEnablePasswordNameCheck: ( 'NAMECHECK', 'no' ),
    getMinDiff: ('MINDIFF', '0' )
}


class SolarisPasswordPolicyPlugin:
    def __init__(self, realplugin, target):
        self.target = target
        self.realplugin = realplugin
        
    def shutdown(self):
        pass

    def _setKeys(self, result, path, keys, apply=True, has_comments=True):
        xpath = self.target.translatePath(path)
        fd = self.target.getFile(xpath)

        ke = KeyValueEditor(java.lang.String(fd), has_comments=has_comments)
        for k in keys:
            current = ke.get(k)
            if current != keys[k]:
                result.diff(path, k, current, keys[k])
                ke.set(k, keys[k])

        if ke.hasChanged() and apply:
            self.target.putFile(xpath, ke.getData(), True)

        return ke.hasChanged()


    def getGlobalPolicy(self, policy):
        g = policy.getResource(PolicyConstants.GLOBAL, "GLOBAL")
        if g is not None and g.getPasswdID() is not None:
            gp = policy.getRule(PolicyConstants.PASSWORD_MANAGEMENT, g.getPasswdID())
            if gp != PasswordManagement(None, None):
                return gp

        return None

    def getUserPolicy(self, policy):
        # find users with attached password policy
        userpolicy = {}
        users = policy.resourceMap.get(PolicyConstants.USER) or {}
        loginpolicy = policy.ruleMap.get(PolicyConstants.PASSWORD_MANAGEMENT) or {}
        for username in users:
            user = users.get(username)
            if user is None or user.getPasswdID() is None:
                continue

            up = loginpolicy.get(user.getPasswdID())
            userpolicy[username] = up

        return userpolicy

    
    def updateShadowRecord(self, record, (result, globalpolicy, userpolicy)):
        """
        Processes a record in /etc/shadow, updating the min field.
        """
        user = record['username']
        result.startAction(User(user, "TAMOS"))

        up = None
        if user in userpolicy:
            up = userpolicy[user]

        value = None
        if up is not None:
            value = up.getMinPasswordDays()
            result.addRule(up)

        if value is None:
            value = globalpolicy.getMinPasswordDays()
            if value is None:
                # no policy to apply
                result.finishAction(npm_s_success, [])
                return

            result.addRule(globalpolicy)

        current = record["min"] or None
        if value != current:
            result.diff(SHADOW, "min", current, value)
            record["min"] = value or ""

        result.finishAction(npm_s_success, [])
        return record


    def applyPasswordPolicy(self, result, globalpolicy, userpolicy, apply):

        # apply global policy
        result.startAction(Global("global", "TAMOS"))
        result.addRule(globalpolicy)

        keys = {}

        for method in mapping:
            (field, default) = mapping[method]
            # get value from policy, applying translation if there is one,
            # and using the default value if none is supplied
            value = method(globalpolicy)
            if value is None:
                value = default

            keys[field] = value

        changed = self._setKeys(result, PASSWD_DEFAULTS, keys, apply)

        result.finishAction(npm_s_success, [])

        # apply user policy (minimum password age)
        shadowpath = self.target.translatePath(SHADOW)
        shadow = SolarisShadowFile(self.target.getFile(shadowpath))

        shadow.update(self.updateShadowRecord, (result, globalpolicy, userpolicy))

        if shadow.hasChanged() and apply:
            self.target.putFile(shadowpath, shadow.getData(), True)
            changed = True

        return changed


    def policyUpdate(self, result, delta, apply):
        gp = self.getGlobalPolicy(delta.newPolicy)
        up = self.getUserPolicy(delta.newPolicy)

        # if there's any applicable policy at all before and after this delta,
        # then we're not going to manage the login activity policy at all.
        if gp is None and up == {}:
            ogp = self.getGlobalPolicy(delta.oldPolicy)
            oup = self.getUserPolicy(delta.oldPolicy)
            if ogp is None and oup == {}:
                # no policy, do nothing
                result.finish(npm_s_success, [])
                return

        if gp is None:
            gp = PasswordManagement("default", "TAMOS")

        changed = self.applyPasswordPolicy(result, gp, up, apply)
        if apply == True or changed == False:
            result.finish(npm_s_success, [])
        else:
            result.finish(npm_s_policy_not_compliant, [])


    def verifyPolicy(self, result, policy, reconcile):

        gp = self.getGlobalPolicy(policy)
        up = self.getUserPolicy(policy)
        if gp is None and up == {}:
            # no policy, so do nothing
            result.finish(npm_s_success, [])
            return

        changed = self.applyPasswordPolicy(result, gp, up, reconcile)
        if changed is False:
            result.finish(npm_s_success, [])
        elif reconcile:
            result.finish(npm_s_policy_reconciled, [])
        else:
            result.finish(npm_s_policy_not_compliant, [])

# vim: set ts=4 sw=4 expandtab :
