#
# IBM Confidential
# OCO Source Materials
# Tivoli Access Manager for Operating Systems - Agent
# TIVO-TAMOSA
#
# (C) Copyright IBM Corp. 2008
# The source code for this program is not published or otherwise
# divested of its trade secrets, irrespective of what has been
# deposited with the U.S. Copyright Office.
#

#
# aix user audit implementation
#

from com.ibm.amos.agent.db import PolicyConstants
from com.ibm.amos.agent.db.resources import User
from com.ibm.amos.agent.db.resources import File
from com.ibm.amos.nls.pdopmidx import *

import java.lang.Exception
import java.lang.String
import java.lang.StringBuffer
import java.util.logging.Level

from ossutil import wrapPyException
from auditutil import determineAuditUsers
from ossutil.aixconfig import AixSystemConfigFile
from ossutil.aixaudit import restartAuditDaemon


AUDIT_CONFIG_FILE = "/etc/security/audit/config"
TAMOS_AUDIT_CLASSES = [ "TAMOS_authzn", "TAMOS_login", "TAMOS_exec", "TAMOS_file"]

tamos_audit_level_map = {
    "authnone"    : None,
    "permit"      : "TAMOS_authzn",
    "deny"        : "TAMOS_authzn",
    "loginpermit" : "TAMOS_login",
    "logindeny"   : "TAMOS_login",
    "tracenone"   : None,
    "exec"        : "TAMOS_exec",
    "exec_l"      : "TAMOS_exec",
    "file"        : "TAMOS_file"
}

TRACENONE_ANTI_LEVELS = [ "exec", "exec_l", "file" ]

tamos_audit_class_map = {
    "TAMOS_authzn" : [ "FS_Chdir",
                       "USER_SU", "PROC_SetUserIDs",
                       "PROC_Kill",
                       "USER_Login",
                       "FILE_Open", "FILE_Link", "FS_Mkdir", "FILE_Symlink",
                       "FILE_Rename",
                       "FILE_Utimes",
                       "FILE_Truncate", "FS_Rmdir",
                       "FILE_Owner", "FILE_Fchown",
                       "FILE_Mode", "FILE_Fchmod",
                       "FILE_Read", "FILE_ReadXacl", "FILE_FReadXacl",
                       "FILE_Write", "FILE_WriteXacl", "FILE_FWriteXacl",
                       "PROC_Execute", "PROC_LPExecute" ],
    "TAMOS_login"  : [ "USER_Login" ],
    "TAMOS_exec"   : [ "PROC_Execute", "PROC_LPExecute" ],
    "TAMOS_file"   : [ "FILE_Open", "FILE_Link", "FS_Mkdir", "FILE_Symlink",
                       "FILE_Rename",
                       "FILE_Utimes",
                       "FILE_Truncate", "FS_Rmdir",
                       "FILE_Owner", "FILE_Fchown",
                       "FILE_Mode", "FILE_Fchmod",
                       "FILE_Read", "FILE_ReadXacl", "FILE_FReadXacl",
                       "FILE_Write", "FILE_WriteXacl", "FILE_FWriteXacl" ]
}
                       

class AixUserAuditPlugin:
    def __init__(self, realplugin, target):
        self.target = target
        self.realplugin = realplugin
        self.__l = target.logger

    def shutdown(self):
        # nothing
        pass


    def combineAuditSettings(self, result, user, user_audit, group_audit):
        """
        Combines the user and group audit levels for a user to produce a single
        audit string. The resulting string can be appended to the user's existing
        audit settings and normalized.
        """

        # shortcut special case for users with 'authnone' audit level
        if user_audit is not None and 'authnone' in user_audit.getLevels():
            self.__l.finer("User %s has 'authnone' audit level => disabling all audit" % user)
            result.addRule(user_audit)
            return []

        # gather audit levels to apply from groups
        apply = {}
        for audit in group_audit:
            result.addRule(audit)
            for l in audit.getLevels():
                apply[l] = 1

        # and from the user, if there is one
        if user_audit is not None:
            result.addRule(user_audit)
            for l in user_audit.getLevels():
                apply[l] = 1

        # Special case - if tracenone is set on the user, remove the audit 
        # levels that it covers
        if user_audit is not None and 'tracenone' in user_audit.getLevels():
            self.__l.finer("User %s has 'tracenone' audit level => disabling any trace audit" % user)
            for l in TRACENONE_ANTI_LEVELS:
                if l in apply:
                    del apply[l]

        # use the map to convert from TAMOS audit settings to AIX audit classes
        classes = {}
        for l in apply:
            if l in tamos_audit_level_map:
                self.__l.finest("Applying audit level '%s'" % l)
                cls = tamos_audit_level_map[l]
                if cls is not None:
                    classes[cls] = 1
            else:
                # Log an error, this should never happen
                pass

        self.__l.finer("Combined audit setting for %s: %s" % (user, ",".join(classes.keys())))

        return classes.keys()

    def applyPolicy(self, result, policy, update, warnOnly = False):
        """
        Apply a policy, either as an update or a reconciliation, possibly
        not even applying the policy.
        """
        outcome = npm_s_success

        try:
            # Read the current state
            xau = self.target.translatePath(AUDIT_CONFIG_FILE)
            data = java.lang.String(self.target.getFile(xau))
            userAuditFile = AixSystemConfigFile(data)
            userAuditFile.parse()

            # If we have to update the class definitions, do that first.
            if not warnOnly:
                outcome = self.setClassDefinitions(result, userAuditFile)
                if outcome != npm_s_success:
                    result.finish(outcome, [])
                    return

                if userAuditFile.hasChanged():
                    result.startAction(File(AUDIT_CONFIG_FILE, None, None))
                    result.newStep("Updating user audit file")
                    self.__l.finest("New config file content:\n%s\n" % (userAuditFile.getText()))
                    self.target.putFile(xau, userAuditFile.getText(), True)
                    # Restart the audit daemon to pick up configuration changes
                    restartAuditDaemon(self.target, result)
		    result.finishAction(npm_s_success, [])

            # Now that we know the TAMOS_* class definitions are available, apply
            # the user audit policy.
            outcome = self.applyPolicyImpl(result, userAuditFile, policy, update, warnOnly)

            result.finish(outcome, [])
            
        except java.lang.Exception, e:
            result.finish(e)
        except Exception, e:
            result.finish(wrapPyException(self.realplugin, e))


    def setClassDefinitions(self, result, userAuditFile):
        """
        Sets the audit class definitions used by TAMOS Agent in the audit file.
        """
        classesStanza = "classes"
        defns = {}
        for auditClass in tamos_audit_class_map.keys():
            defns[auditClass] = 1

        def updateClassDefnCb(stanzaName, name, values, data):
            (result, defns) = data
            retval = values

            if tamos_audit_class_map.has_key(name):
                # It's one of ours. Check if it has all the appropriate values
                # set
                expected = tamos_audit_class_map[name]

                surplus = [ v for v in values if v not in expected ]
                missing = [ v for v in expected if v not in values ]

                if len(surplus) > 0 or len(missing) > 0:
                    retval = expected

                del defns[name]

            return retval

        userAuditFile.update(classesStanza, updateClassDefnCb, (result, defns))

        count = userAuditFile.getStanzaEntryCount(classesStanza)
        # AIX imposes a hard limit of 31 user-defined classes
        if (count + len(defns)) > 31:
            # We can't set our class definitions. We can't proceed.
            return npm_s_aix_audit_class_limit
        
        for auditClass in defns.keys():
            userAuditFile.appendEntry(classesStanza, auditClass, tamos_audit_class_map[auditClass])

        return npm_s_success

    def runChuser(self, username, auditValues):
        """
        Run chuser to update the user's audit classes
        """
        if auditValues is None:
            auditValues = [""]

        args = ["/usr/bin/chuser", "auditclasses=" + ",".join(auditValues), username]
        out = java.lang.StringBuffer()
        err = java.lang.StringBuffer()
        rc = self.target.execute(args, None, out, err)
        if rc != 0:
            self.__l.warning("Received error code %d executing '%s': %s" % (rc, " ".join(args), err))

    def determineAuditClasses(self, result, userAuditFile):
        map = {}

        def getAuditClassCb(stanzaName, name, values, data):
            (result, map) = data
            map[name] = 1
            return values
        
        userAuditFile.update("classes", getAuditClassCb, (result, map))

        return map
        

    def applyPolicyImpl(self, result, userAuditFile, policy, update, warnOnly = False):
        """
        Implementation of actual policy application. This has been implemented
        as a separate function mainly for ease of unit testing.
        """
        outcome = npm_s_success

        # Find the user audit classes that are available in the file.
        auditClasses = self.determineAuditClasses(result, userAuditFile)

        # Find the users that have user audit applied to them
        (users, haveExec) = determineAuditUsers(self.target, policy)

        # Remove any users that are not in the user registry, otherwise the audit
        # daemon will fail to start
        for user in users:
            out = java.lang.StringBuffer()
            err = java.lang.StringBuffer()
            rc = self.target.execute(("/usr/sbin/lsuser", user), None, out, err)
            if rc != 0:
                # Probably means the user doesn't exist
                self.__l.finer("Ignoring user %s because they are not known on this system." % (user))
                del users[user]

        # Callback function to update the existing entries in the users
        # stanza. This will remove any existing TAMOS agent audit classes
        # and replace them with what is configured in the policy.
        # It will also have the effect of removing entries where there are
        # no audit settings defined for that user.
        def updateAuditCb(stanzaName, name, values, data):
            (result, users, audit_classes) = data
            result.startAction(User(name, "TAMOS"))
            
            retval = values
            
            tamosValues    = [ v for v in values if v in TAMOS_AUDIT_CLASSES ]
            nonTamosValues = [ v for v in values if v not in TAMOS_AUDIT_CLASSES ]
            
            self.__l.finer("Processing existing user %s with current audit settings: %s" % (name, ",".join(values)))
            
            if name in users:
                # They have a TAMOS user audit policy set, so check that
                # what they have is what we expect
                audit_settings = users[name]
                expTamosValues = self.combineAuditSettings(result, name, audit_settings[0], audit_settings[1:])
                
                self.__l.finer("Expect user to have the TAMOS audit classes: %s" % (",".join(expTamosValues)))
                
                surplus = [ v for v in tamosValues if v not in expTamosValues ]
                missing = [ v for v in expTamosValues if v not in tamosValues ]
                
                if len(surplus) > 0 or len(missing) > 0:
                    # We have some work to do
                    retval = [ v for v in nonTamosValues if v in audit_classes]
                    retval.extend([ v for v in expTamosValues if v in audit_classes])
            else:
                # They do not have a TAMOS user audit policy set, so we
                # expect them to have no TAMOS audit classes set
                self.__l.finer("Expect user to have no TAMOS audit classes.")
                if len(tamosValues) > 0:
                    retval = [v for v in nonTamosValues if v in audit_classes]

            # Now the fun stuff, reporting this back in a myriad of
            # different ways.
            if not retval == values or name in users:
                result.diff("user", name, ",".join(values), ",".join(retval))
            if retval == values:
                result.finishAction(npm_s_success, [])
            elif update:
                self.runChuser(name, retval)
                result.finishAction(npm_s_success, [])
            elif warnOnly:
                result.finishAction(npm_s_policy_not_compliant, [])
                retval = values
            else:
                self.runChuser(name, retval)
                result.finishAction(npm_s_policy_reconciled, [])
                    
            # We've processed this user, forget about them
            if name in users:
                del users[name]

            return retval

        # Use the callback to make the update
        userAuditFile.update("users", updateAuditCb, (result, users, auditClasses))

        # Append any that were not previously existing in the audit file
        self.__l.finer("Adding any remaining users with no existing audit settings")
        for user in users:
            result.startAction(User(user, "TAMOS"))
            auditSettings = users[user]
            tamosValues = self.combineAuditSettings(result, user, auditSettings[0], auditSettings[1:])
            setValues = [ v for v in tamosValues if v in auditClasses ]

            if tamosValues:
                result.diff("user", user, None, ",".join(setValues))
                if update:
                    self.runChuser(user, setValues)
                    result.finishAction(npm_s_success, [])
                else:
                    if warnOnly:
                        result.finishAction(npm_s_policy_not_compliant, [])
                    else:
                        self.runChuser(user, setValues)
                        result.finishAction(npm_s_policy_reconciled, [])
            else:
                # If they don't have a tamosValue, it means they have the 'none'
                # audit setting set. And they're not in the file. Can't get
                # much more none than that.
                result.finishAction(npm_s_success, [])
                
        return outcome

    def policyUpdate(self, result, delta, apply):
        """
        Apply a policy update to the user audit settings.
        However, because we have to manually expand the groups, we apply the
        full policy every time, since the delta may not actually pick up all
        the changes.
        """
        self.applyPolicy(result, delta.newPolicy, True, False)


    def verifyPolicy(self, result, policy, reconcile):
        """
        Verify that the policy that is currently on the system complies with
        what is set in TAMOS. Optionally, update the policy to reflect with
        what TAMOS has.
        """
        self.applyPolicy(result, policy, False, not reconcile)


# vim: set ts=4 sw=4 expandtab :
