#
# IBM Confidential
# OCO Source Materials
# Tivoli Access Manager for Operating Systems - Agent
# TIVO-TAMOSA
#
# (C) Copyright IBM Corp. 2008
# The source code for this program is not published or otherwise
# divested of its trade secrets, irrespective of what has been
# deposited with the U.S. Copyright Office.
#

#
# solaris user audit implementation
#

from __future__ import generators

from com.ibm.amos.agent.db import PolicyConstants
from com.ibm.amos.agent.db.resources import User
from com.ibm.amos.nls.pdopmidx import *

import java.lang.Exception
import java.lang.String
import java.lang.StringBuffer
import java.util.logging.Level

from ossutil import wrapPyException
from ossutil.solarisaudit import *
from auditutil import determineAuditUsers, auditPolicyExists
from ossutil.configeditor import LineEditor

def modifierify(modifier, flags):
    """
    Constructs an audit flag string applying a particular modifier to a
    list of classes.
    """
    return modifier + ("," + modifier).join(flags)

def smash(audit_a, audit_b):
    """
    Smashes two audit flag strings together.
    """
    if audit_a == "":
        return audit_b
    elif audit_b == "":
        return audit_a
    else:
        return audit_a + "," + audit_b

# whenever we're enabling 'all' audit events, these are the classes we enable:
ALL_AUDIT_CLASSES = ['fr','fw','fa','fm','fc','fd','lo','ex','pm']
# if we only want file-related events, these are the classes we enable:
ALL_FILE_AUDIT_CLASSES = ['fr','fw','fa','fc','fd','fm']


# default 'always' and 'never' audit policy: turn off all the classes we're
# controlling, and remove them from the 'never' setting, allowing the global
# audit policy to apply.
EXCLUDE_ALL_AUDIT_CLASSES = modifierify("^", ALL_AUDIT_CLASSES)
ENABLE_ALL_AUDIT_CLASSES = modifierify("", ALL_AUDIT_CLASSES)

# mappings from TAMOS audit settings to default Solaris audit classes in (always, never) pairs
# this doesn't seem to be clever enough; there's some interaction between auditauth and
# audittrace flags that isn't captured here.
tamos_audit_level_map = {
    "authnone": ["", "" ],                  # special handling at user level
    "permit": [modifierify("+", ALL_AUDIT_CLASSES), ""],
    "deny": [modifierify("-", ALL_AUDIT_CLASSES), ""],
    "loginpermit": ["+lo", ""],
    "logindeny": ["-lo", ""],
    "exec": ["ex", ""],
    "exec_l": ["ex", ""],
    "file": [modifierify("", ALL_FILE_AUDIT_CLASSES), ""],
    "tracenone": ["", modifierify("", ALL_FILE_AUDIT_CLASSES)]         # unsure about this one.
}


                
class SolarisUserAuditFile:
    """
    Specialised editor for the Solaris audit_user file
    """

    def __init__(self, data):
        self.edit = LineEditor(data, has_comments = True, has_continuations = True)

    def entries(self):
        """
        Generator that returns user, always-audit, never-audit
        tuples for each user audit setting in the file.
        """
        for line in self.edit.lines():
            try:
                (user, always, never) = line.split(":")
                yield (user, always, never)
            except ValueError:
                # malformed line; ignore it
                # probably log something
                continue

    def update(self, callback, data):
        """
        Calls the callback for each entry in turn.  The callback can
        return None to remove the entry, or a tuple containing new
        always-audit and never-audit settings for the user to update
        the entry.
        """

        def cb(line, (callback, cbdata)):
            try:
                (user, always, never) = line.split(":")
                data = callback(user, always, never, cbdata)
                if data is None:
                    return None
                else:
                    (always, never) = data
                    if always == "" and never == "":
                        # filter out empty entries too
                        return None
                    return "%s:%s:%s" % (user, always, never)
            except ValueError:
                # file is malformed.  just ignore this line.
                # probably log something.
                return line

        self.edit.update(cb, (callback, data))

    def append(self, user, always, never):
        self.edit.append("%s:%s:%s" % (user, always, never))
    
    def getData(self):
        return self.edit.data

    def changed(self):
        return self.edit.hasChanged()




class SolarisUserAuditPlugin:
    def __init__(self, realplugin, target):
        self.target = target
        self.realplugin = realplugin
        self.__l = target.logger
        # probably check bsm is enabled?
        # maybe check for NIS/LDAP/whatever, complain if nsswitch goes there before files,
        # complain bitterly if it doesn't go to files at all
 
    def shutdown(self):
        # nothing
        pass

    def combineAuditSettings(self, result, user, user_audit, group_audit):
        """
        Combines the user and group audit levels for a user to produce a single
        audit flag string.  The resulting string can be appended to the user's
        existing audit settings and normalized.
        """

        # shortcut special case for users with 'authnone' audit level
        if user_audit is not None and 'authnone' in user_audit.getLevels():
            self.__l.finer("User %s has 'authnone' audit level => disabling all audit" % user)
            result.addRule(user_audit)
            return (EXCLUDE_ALL_AUDIT_CLASSES, ENABLE_ALL_AUDIT_CLASSES)

        always = EXCLUDE_ALL_AUDIT_CLASSES
        never = EXCLUDE_ALL_AUDIT_CLASSES
        apply = {}

        # gather audit levels to apply from grops
        for audit in group_audit:
            result.addRule(audit)
            for l in audit.getLevels():
                apply[l] = 1

        # and from the user audit setting, if there is one
        if user_audit is not None:
            result.addRule(user_audit)
            for l in user_audit.getLevels():
                apply[l] = 1

        # apply them all; just append the always and never flag strings for
        # each audit level to the default flag string.  they'll be normalized
        # later on.
        for l in apply:
            if l in tamos_audit_level_map:
                self.__l.finest("Applying audit level '%s'" % l)
                (lalways, lnever) = tamos_audit_level_map[l]
                always = smash(always, lalways)
                never = smash(never, lnever)
            else:
                # probably log something; this shouldn't ever happen
                pass

        self.__l.finer("Combined audit settings for %s: [%s,%s]" % (user, always, never))
        return (always, never)




    def policyUpdate(self, result, delta, apply):

        if (auditPolicyExists(delta.newPolicy) is False) and (auditPolicyExists(delta.oldPolicy) is False):
            self.__l.finer("No audit policy exists")
            result.finish(npm_s_success, [])
            return

        try:
            outcome = npm_s_success

            # represent the whole thing as an action on the audit_user file
            (users, haveExec) = determineAuditUsers(self.target, delta.newPolicy)

            # read current user audit settings
            xau = self.target.translatePath(AUDIT_USER)
            data = java.lang.String(self.target.getFile(xau))
            userAuditFile = SolarisUserAuditFile(data)

            # update existing settings for users already present in the file
            def update_cb(user, always, never, data):
                (result, users) = data

                result.startAction(User(user, "TAMOS"))
                # if the user's audit settings are not specified by the policy,
                # then disable all the audit classes we control
                if user in users:
                    self.__l.finer("Processing existing audit settings for user %s: %s, %s" % (user, always, never))
                    audit_settings = users[user]
                    (policyalways, policynever) = self.combineAuditSettings(result, user, audit_settings[0], audit_settings[1:])

                    # forget about this user
                    del users[user]
                else:
                    self.__l.finer("No audit policy covers user %s; existing settings: %s, %s" % (user, always, never))
                    policyalways = EXCLUDE_ALL_AUDIT_CLASSES
                    policynever = EXCLUDE_ALL_AUDIT_CLASSES

                # apply policy to current settings and return
                # if the user settings simplify down to nothing, then the audit file editor
                # will remove the entry for this user.
                newalways = normalizeAuditFlags(smash(always, policyalways), audit_metaclass_map)
                newnever = normalizeAuditFlags(smash(never, policynever), audit_metaclass_map)
                if newalways != always or newnever != never:
                    result.diff(AUDIT_USER, user, "%s:%s" % (always, never), "%s:%s" % (newalways, newnever))
                result.finishAction(npm_s_success, [])
                return (newalways, newnever)

            # update user audit settings that already exist in the file
            self.__l.fine("Updating existing audit settings from audit_user file")
            userAuditFile.update(update_cb, (result, users))

            # and append any that are missing
            self.__l.fine("Adding any remaining users with no existing audit settings")
            for user in users:
                result.startAction(User(user, "TAMOS"))
                auditSettings = users[user]
                (always, never) = self.combineAuditSettings(result, user, auditSettings[0], auditSettings[1:])

                always = normalizeAuditFlags(always, audit_metaclass_map)
                never = normalizeAuditFlags(never, audit_metaclass_map)
                
                if always != "" or never != "":
                    result.diff(AUDIT_USER, user, None, "%s:%s" % (always, never))
                    userAuditFile.append(user, always, never)

                result.finishAction(npm_s_success, [])

            if haveExec:
                #result.newStep("Enabling auditing of process arguments")
                # if any trace-exec settings are present, auditconfig -setpolicy +argv
                # and/or add such a line to /etc/security/audit_startup
                pass
     
            # update the audit_user file if we've changed anything and we're being asked to apply the policy
            if apply and userAuditFile.changed():
                self.__l.finest("New audit_user content:\n%s\n" % userAuditFile.getData())
                self.target.putFile(xau, userAuditFile.getData(), True)

            result.finish(outcome, [])
        except java.lang.Exception, e:
            result.finish(e)
        except Exception, e:
            result.finish(wrapPyException(self.realplugin, e))

        
    def verifyPolicy(self, result, policy, reconcile):

        if auditPolicyExists(policy) is False:
            self.__l.finer("No audit policy exists")
            result.finish(npm_s_success, [])
            return

        outcome = npm_s_success
        try:
            (users, haveExec) = determineAuditUsers(self.target, policy)

            # read current user audit settings
            xau = self.target.translatePath(AUDIT_USER)
            data = java.lang.String(self.target.getFile(xau))
            userAuditFile = SolarisUserAuditFile(data)

            def update_cb(user, always, never, data):
                (result, users, reconcile) = data

                result.startAction(User(user, "TAMOS"))
                self.__l.finer("Verifying audit settings for user %s; current settings: %s, %s" % (user, always, never))

                # calculate the audit flag strings specified by the policy for this user
                if user in users:
                    audit_settings = users[user]
                    (policyalways, policynever) = self.combineAuditSettings(result, user, audit_settings[0], audit_settings[1:])
                    del users[user]
                else:
                    policyalways = EXCLUDE_ALL_AUDIT_CLASSES
                    policynever = EXCLUDE_ALL_AUDIT_CLASSES

                # if the user's audit settings already comply with the policy, then combining them with
                # the policy audit settings won't change them.
                def calculate_expected(flags, policyflags):
                    normflags = normalizeAuditFlags(flags, audit_metaclass_map)
                    expflags = normalizeAuditFlags(smash(flags, policyflags), audit_metaclass_map)
                    return (normflags, expflags)

                (normalways, expalways) = calculate_expected(always, policyalways)
                (normnever, expnever) = calculate_expected(never, policynever)

                if (normalways, normnever) != (expalways, expnever):
                    result.diff(AUDIT_USER, user, "%s:%s" % (always, never), "%s:%s" % (expalways, expnever))
                    if reconcile:
                        (always, never) = (expalways, expnever)
                        result.finishAction(npm_s_policy_reconciled, [])
                    else:
                        result.finishAction(npm_s_policy_not_compliant, [])
                else:
                    result.finishAction(npm_s_success, [])

                return (always, never)

            userAuditFile.update(update_cb, (result, users, reconcile))

            # fix up any users left in the map
            for user in users:

                result.startAction(User(user, "TAMOS"))

                audit_settings = users[user]
                (always, never) = self.combineAuditSettings(result, user, audit_settings[0], audit_settings[1:])
                always = normalizeAuditFlags(always, audit_metaclass_map)
                never = normalizeAuditFlags(never, audit_metaclass_map)

                if always == "" and never == "":
                    # user isn't really covered by policy, so don't do anything
                    result.finishAction(npm_s_success, [])
                else:
                    result.diff(AUDIT_USER, user, None, "%s:%s" % (always, never))
                    if reconcile:
                        userAuditFile.append(user, always, never)
                        result.finishAction(npm_s_policy_reconciled, [])
                    else:
                        result.finishAction(npm_s_policy_not_compliant, [])

            if userAuditFile.changed():
                if reconcile:
                    self.__l.finest("New audit_user content:\n%s\n" % userAuditFile.getData())
                    self.target.putFile(xau, userAuditFile.getData(), True)
                    outcome = npm_s_policy_reconciled
                else:
                    outcome = npm_s_policy_not_compliant

            result.finish(outcome, [])
        except java.lang.Exception, e:
            result.finish(e)
        except Exception, e:
            result.finish(wrapPyException(self.realplugin, e))
        

# vim: set ts=4 sw=4 expandtab :
